#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: Contributors to the Ansible project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import annotations

DOCUMENTATION = r"""
---
module: s3_bucket
version_added: 1.0.0
short_description: Manage S3 buckets in AWS, DigitalOcean, Ceph, Walrus, FakeS3 and StorageGRID
description:
  - Manage S3 buckets.
  - Compatible with AWS, DigitalOcean, Ceph, Walrus, FakeS3 and StorageGRID.
  - When using non-AWS services, O(endpoint_url) should be specified.
author:
  - Rob White (@wimnat)
  - Aubin Bikouo (@abikouo)
options:
  force:
    description:
      - When trying to delete a bucket, delete all keys (including versions and delete markers)
        in the bucket first (an S3 bucket must be empty for a successful deletion).
    type: bool
    default: false
  name:
    description:
      - Name of the S3 bucket.
    required: true
    type: str
  policy:
    description:
      - The JSON policy as a string. Set to the string V("null") to force the absence of a policy.
    type: json
  ceph:
    description:
      - Enable API compatibility with Ceph RGW.
      - It takes into account the S3 API subset working with Ceph in order to provide the same module
        behaviour where possible.
      - Requires O(endpoint_url) if O(ceph=true).
    aliases: ['rgw']
    type: bool
    default: false
  account_regional:
    description:
      - Whether to create the bucket in the account-regional namespace instead of the global namespace.
      - When V(true), the bucket name must follow the format C(prefix-accountId-region-an).
      - Can only be set during bucket creation. Cannot be changed for existing buckets.
      - Not supported for directory buckets.
      - Requires botocore >= 1.42.67
    type: bool
    version_added: 11.3.0
  requester_pays:
    description:
      - With Requester Pays buckets, the requester instead of the bucket owner pays the cost
        of the request and the data download from the bucket.
    type: bool
  state:
    description:
      - Create or remove the S3 bucket.
    required: false
    default: present
    choices: [ 'present', 'absent' ]
    type: str
  versioning:
    description:
      - Whether versioning is enabled or disabled (note that once versioning is enabled, it can only be suspended).
    type: bool
  encryption:
    description:
      - Describes the default server-side encryption to apply to new objects in the bucket.
        In order to remove the server-side encryption, the encryption needs to be set to 'none' explicitly.
      - "Note: Since January 2023 Amazon S3 doesn't support disabling encryption on S3 buckets."
    choices: [ 'none', 'AES256', 'aws:kms' ]
    type: str
  encryption_key_id:
    description:
      - KMS master key ID to use for the default encryption.
      - If not specified then it will default to the AWS provided KMS key.
      - This parameter is only supported if O(encryption) is V(aws:kms).
    type: str
  bucket_key_enabled:
    description:
      - Enable S3 Bucket Keys for SSE-KMS on new objects.
      - See the AWS documentation for more information
        U(https://docs.aws.amazon.com/AmazonS3/latest/userguide/bucket-key.html).
      - Bucket Key encryption is only supported if O(encryption=aws:kms).
    required: false
    type: bool
    version_added: 4.1.0
  public_access:
    description:
      - Configure public access block for S3 bucket.
      - This option cannot be used together with O(delete_public_access).
      - |
        Note: At the end of April 2023 Amazon updated the default settings to block public access by
        default.  While the defaults for this module remain unchanged, it is necessary to explicitly
        pass the O(public_access) parameter to enable public access ACLs.
    suboptions:
      block_public_acls:
        description: Sets BlockPublicAcls value.
        type: bool
        default: false
      block_public_policy:
        description: Sets BlockPublicPolicy value.
        type: bool
        default: false
      ignore_public_acls:
        description: Sets IgnorePublicAcls value.
        type: bool
        default: false
      restrict_public_buckets:
        description: Sets RestrictPublicAcls value.
        type: bool
        default: false
    type: dict
    version_added: 1.3.0
  delete_public_access:
    description:
      - Delete public access block configuration from bucket.
      - This option cannot be used together with a O(public_access) definition.
    default: false
    type: bool
    version_added: 1.3.0
  object_ownership:
    description:
      - Allow bucket's ownership controls.
      - V(BucketOwnerEnforced) - ACLs are disabled and no longer affect access permissions to your
        bucket. Requests to set or update ACLs fail. However, requests to read ACLs are supported.
        Bucket owner has full ownership and control. Object writer no longer has full ownership and
        control.
      - V(BucketOwnerPreferred) - Objects uploaded to the bucket change ownership to the bucket owner
        if the objects are uploaded with the bucket-owner-full-control canned ACL.
      - V(ObjectWriter) - The uploading account will own the object
        if the object is uploaded with the bucket-owner-full-control canned ACL.
      - This option cannot be used together with a O(delete_object_ownership) definition.
      - V(BucketOwnerEnforced) has been added in version 3.2.0.
      - "Note: At the end of April 2023 Amazon updated the default setting to V(BucketOwnerEnforced)."
    choices: [ 'BucketOwnerEnforced', 'BucketOwnerPreferred', 'ObjectWriter' ]
    type: str
    version_added: 2.0.0
  object_lock_enabled:
    description:
      - Whether S3 Object Lock to be enabled.
      - Defaults to V(false) when creating a new bucket.
    type: bool
    version_added: 5.3.0
  delete_object_ownership:
    description:
      - Delete bucket's ownership controls.
      - This option cannot be used together with a O(object_ownership) definition.
    default: false
    type: bool
    version_added: 2.0.0
  acl:
    description:
      - The canned ACL to apply to the bucket.
      - If your bucket uses the bucket owner enforced setting for S3 Object Ownership,
        ACLs are disabled and no longer affect permissions.
      - If O(public_access.ignore_public_acls=true) setting O(acl) to a value other than V(private)
        may return success but not have an affect.
    choices: [ 'private', 'public-read', 'public-read-write', 'authenticated-read' ]
    type: str
    version_added: 3.1.0
  validate_bucket_name:
    description:
      - Whether the bucket name should be validated to conform to AWS S3 naming rules.
      - On by default, this may be disabled for S3 backends that do not enforce these rules.
      - See U(https://docs.aws.amazon.com/AmazonS3/latest/userguide/bucketnamingrules.html).
    type: bool
    version_added: 3.1.0
    default: true
  dualstack:
    description:
      - Enables Amazon S3 Dual-Stack Endpoints, allowing S3 communications using both IPv4 and IPv6.
      - Mutually exclusive with O(endpoint_url).
    type: bool
    default: false
    version_added: 6.0.0
  accelerate_enabled:
    description:
      - Enables Amazon S3 Transfer Acceleration, sent data will be routed to Amazon S3 over an optimized network path.
      - Transfer Acceleration is not available in AWS GovCloud (US).
      - See U(https://docs.aws.amazon.com/govcloud-us/latest/UserGuide/govcloud-s3.html#govcloud-S3-diffs).
    type: bool
    version_added: 8.1.0
  object_lock_default_retention:
    description:
      - Default Object Lock configuration that will be applied by default to
        every new object placed in the specified bucket.
      - O(object_lock_enabled) must be included and set to V(True).
      - Object lock retention policy can't be removed.
    suboptions:
      mode:
        description: Type of retention modes.
        choices: [ "GOVERNANCE", "COMPLIANCE"]
        required: true
        type: str
      days:
        description:
            - The number of days that you want to specify for the default retention period.
            - Mutually exclusive with O(object_lock_default_retention.years).
        type: int
      years:
        description:
            - The number of years that you want to specify for the default retention period.
            - Mutually exclusive with O(object_lock_default_retention.days).
        type: int
    type: dict
    version_added: 8.1.0
  inventory:
    description:
      - Enable S3 Inventory, saving list of the objects and their corresponding
        metadata on a daily or weekly basis for an S3 bucket.
    type: list
    elements: dict
    suboptions:
      destination:
        description: Contains information about where to publish the inventory results.
        type: dict
        required: True
        suboptions:
          account_id:
            description: The account ID that owns the destination S3 bucket. If no account ID is provided, the owner is not validated before exporting data.
            type: str
          bucket:
            description: The Amazon Resource Name (ARN) of the bucket where inventory results will be published.
            type: str
            required: True
          format:
            description: Specifies the output format of the inventory results.
            type: str
            choices: [ 'CSV', 'ORC', 'Parquet' ]
            required: True
          prefix:
            description: The prefix that is prepended to all inventory results.
            type: str
      filter:
        description: The prefix that an object must have to be included in the inventory results.
        type: str
      id:
        description: The ID used to identify the inventory configuration.
        type: str
        required: True
      schedule:
        description: Specifies the schedule for generating inventory results.
        type: str
        choices: [ 'Daily', 'Weekly' ]
        required: True
      included_object_versions:
        description: |
            Object versions to include in the inventory list. If set to All, the list includes all the object versions,
            which adds the version-related fields VersionId, IsLatest, and DeleteMarker to the list. If set to Current,
            the list does not contain these version-related fields.
        type: str
        required: True
        choices: [ 'All', 'Current' ]
      optional_fields:
        description: Contains the optional fields that are included in the inventory results.
        type: list
        elements: str
        choices: [ "Size", "LastModifiedDate", "StorageClass", "ETag",
            "IsMultipartUploaded", "ReplicationStatus", "EncryptionStatus",
            "ObjectLockRetainUntilDate", "ObjectLockMode",
            "ObjectLockLegalHoldStatus", "IntelligentTieringAccessTier",
            "BucketKeyStatus", "ChecksumAlgorithm", "ObjectAccessControlList",
            "ObjectOwner" ]
  logging:
    description:
      - Configure S3 bucket logging to track requests for access to the bucket.
      - To disable logging, explicitly set O(logging.target_bucket) to an empty string.
      - "Note: The target bucket must have appropriate permissions configured. This module does NOT automatically
        configure ACLs on the target bucket. See AWS documentation for required permissions."
    type: dict
    suboptions:
      target_bucket:
        description: The name of the bucket where Amazon S3 should store server access log files.
        type: str
        required: true
      target_prefix:
        description: A prefix for all log object keys.
        type: str
        default: ""
    version_added: 11.3.0
extends_documentation_fragment:
  - amazon.aws.common.modules
  - amazon.aws.region.modules
  - amazon.aws.tags
  - amazon.aws.boto3

notes:
  - If C(requestPayment), C(policy), C(tagging) or C(versioning)
    operations/API aren't implemented by the endpoint, module doesn't fail
    if each parameter satisfies the following condition.
    O(requester_pays) is V(false), O(policy), O(tags), and O(versioning) are V(None).
  - For Walrus O(endpoint_url) should be set to the FQDN of the endpoint with neither scheme nor path.
"""

EXAMPLES = r"""
# Note: These examples do not set authentication details, see the AWS Guide for details.

# Create a simple S3 bucket
- amazon.aws.s3_bucket:
    name: mys3bucket
    state: present

# Create a simple S3 bucket on Ceph Rados Gateway
- amazon.aws.s3_bucket:
    name: mys3bucket
    endpoint_url: http://your-ceph-rados-gateway-server.xxx
    ceph: true

# Create a bucket with account-regional namespace
- amazon.aws.s3_bucket:
    name: my-bucket-111122223333-us-east-1-an
    state: present
    account_regional: true
# Remove an S3 bucket and any keys it contains
- amazon.aws.s3_bucket:
    name: mys3bucket
    state: absent
    force: true

# Create a bucket, add a policy from a file, enable requester pays, enable versioning and tag
- amazon.aws.s3_bucket:
    name: mys3bucket
    policy: "{{ lookup('file','policy.json') }}"
    requester_pays: true
    versioning: true
    tags:
      example: tag1
      another: tag2

# Create a simple DigitalOcean Spaces bucket using their provided regional endpoint
- amazon.aws.s3_bucket:
    name: mydobucket
    endpoint_url: 'https://nyc3.digitaloceanspaces.com'

# Create a bucket with AES256 encryption
- amazon.aws.s3_bucket:
    name: mys3bucket
    state: present
    encryption: "AES256"

# Create a bucket with aws:kms encryption, KMS key
- amazon.aws.s3_bucket:
    name: mys3bucket
    state: present
    encryption: "aws:kms"
    encryption_key_id: "arn:aws:kms:us-east-1:1234/5678example"

# Create a bucket with aws:kms encryption, Bucket key
- amazon.aws.s3_bucket:
    name: mys3bucket
    bucket_key_enabled: true
    encryption: "aws:kms"

# Create a bucket with aws:kms encryption, default key
- amazon.aws.s3_bucket:
    name: mys3bucket
    state: present
    encryption: "aws:kms"

# Create a bucket with public policy block configuration
- amazon.aws.s3_bucket:
    name: mys3bucket
    state: present
    public_access:
      block_public_acls: true
      ignore_public_acls: true
      ## keys == 'false' can be omitted, undefined keys defaults to 'false'
      # block_public_policy: false
      # restrict_public_buckets: false

# Delete public policy block from bucket
- amazon.aws.s3_bucket:
    name: mys3bucket
    state: present
    delete_public_access: true

# Create a bucket with object ownership controls set to ObjectWriter
- amazon.aws.s3_bucket:
    name: mys3bucket
    state: present
    object_ownership: ObjectWriter

# Delete onwership controls from bucket
- amazon.aws.s3_bucket:
    name: mys3bucket
    state: present
    delete_object_ownership: true

# Delete a bucket policy from bucket
- amazon.aws.s3_bucket:
    name: mys3bucket
    state: present
    policy: "null"

# This example grants public-read to everyone on bucket using ACL
- amazon.aws.s3_bucket:
    name: mys3bucket
    state: present
    acl: public-read

# Enable transfer acceleration
- amazon.aws.s3_bucket:
    name: mys3bucket
    state: present
    accelerate_enabled: true

# Default Object Lock retention
- amazon.aws.s3_bucket:
    name: mys3bucket
    state: present
    object_lock_enabled: true
    object_lock_default_retention:
      mode: governance
      days: 1
# Bucket with inventory configuration:
- amazon.aws.s3_bucket:
    name: mys3bucket
    state: present
    inventory:
      - id: mys3bucket-inventory-id
        destination:
          bucket: "arn:aws:s3:::mys3inventorybucket"
        optional_fields:
          - "Size"
        included_object_versions: "All"
        schedule: "Weekly"

# Enable bucket logging
- amazon.aws.s3_bucket:
    name: mys3bucket
    state: present
    logging:
      target_bucket: my-logging-bucket
      target_prefix: logs/mys3bucket/

# Disable bucket logging
- amazon.aws.s3_bucket:
    name: mys3bucket
    state: present
    logging:
      target_bucket: ""
"""

RETURN = r"""
encryption:
    description: Server-side encryption of the objects in the S3 bucket.
    type: dict
    returned: when O(state=present)
    sample: {
                "SSEAlgorithm": "AES256"
            }
name:
    description: Bucket name.
    returned: when O(state=present)
    type: str
    sample: "a-testing-bucket-name"
logging:
    description: S3 bucket's logging configuration.
    type: dict
    returned: when O(state=present)
    sample: {
        "target_bucket": "my-logging-bucket",
        "target_prefix": "logs/"
    }
object_ownership:
    description: S3 bucket's ownership controls.
    type: str
    returned: when O(state=present)
    sample: "BucketOwnerPreferred"
object_lock_default_retention:
    description: S3 bucket's object lock retention policy.
    type: dict
    returned: when O(state=present)
    sample: {
        "Days": 1,
        "Mode": "GOVERNANCE",
        "Years": 0,
    }
policy:
    description: S3 bucket's policy.
    type: dict
    returned: when O(state=present)
    sample: {
        "Statement": [
            {
                "Action": "s3:GetObject",
                "Effect": "Allow",
                "Principal": "*",
                "Resource": "arn:aws:s3:::2d3ce10a8210d36d6b4d23b822892074complex/*",
                "Sid": "AddPerm"
            }
        ],
        "Version": "2012-10-17"
    }
requester_pays:
    description: Indicates that the requester was successfully charged for the request.
    type: bool
    returned: when O(state=present)
    sample: true
tags:
    description: S3 bucket's tags.
    type: dict
    returned: when O(state=present)
    sample: {
        "Tag1": "tag1",
        "Tag2": "tag2"
    }
versioning:
    description: S3 bucket's versioning configuration.
    type: dict
    returned: when O(state=present)
    sample: {
        "MfaDelete": "Disabled",
        "Versioning": "Enabled"
    }
    contains:
        MfaDelete:
            description: Specifies whether MFA delete is enabled in the bucket versioning configuration.
            returned: when O(state=presnet) and MfaDelete configured on bucket.
            type: str
        Versioning:
            description: The versioning state of the bucket.
            type: str
            returned: always
acl:
    description: S3 bucket's canned ACL.
    type: dict
    returned: when O(state=present).
    sample: "public-read"
object_lock_enabled:
    description: Whether S3 Object Lock is enabled.
    type: bool
    returned: when O(state=present)
    sample: false
public_access_block:
    description: Bucket public access block configuration.
    returned: when O(state=present)
    type: dict
    sample: {
                "PublicAccessBlockConfiguration": {
                    "BlockPublicAcls": true,
                    "BlockPublicPolicy": true,
                    "IgnorePublicAcls": true,
                    "RestrictPublicBuckets": true
                }
            }
    contains:
        PublicAccessBlockConfiguration:
            description: The PublicAccessBlock configuration currently in effect for this Amazon S3 bucket.
            type: dict
            contains:
                BlockPublicAcls:
                    description: Specifies whether Amazon S3 should block public access control lists (ACLs) for this bucket and objects in this bucket.
                    type: bool
                BlockPublicPolicy:
                    description: Specifies whether Amazon S3 should block public bucket policies for this bucket.
                    type: bool
                IgnorePublicAcls:
                    description: Specifies whether Amazon S3 should ignore public ACLs for this bucket and objects in this bucket.
                    type: bool
                RestrictPublicBuckets:
                    description: Specifies whether Amazon S3 should restrict public bucket policies for this bucket.
                    type: bool
accelerate_enabled:
    description: S3 bucket acceleration status.
    type: bool
    returned: O(state=present)
    sample: true
bucket_inventory:
    description: S3 bucket inventory configuration.
    type: list
    returned: when O(state=present)
    sample: [
        {
            "IsEnabled": true,
            "Id": "9c2a337ba5fd64de777f499441f83093-inventory-target",
            "Destination": {
                "S3BucketDestination": {
                    "Bucket": "arn:aws:s3:::9c2a337ba5fd64de777f499441f83093-inventory-target",
                    "Format": "CSV"
                    }
                },
            "IncludedObjectVersions": "All",
            "Schedule": {
                "Frequency": "Daily"
            },
            "OptionalFields": []
        }
        ]
"""

import json
import time
import typing
from typing import Iterator
from typing import List
from typing import NoReturn
from typing import Optional
from typing import Tuple
from typing import cast

if typing.TYPE_CHECKING:
    from ansible_collections.amazon.aws.plugins.module_utils.botocore import ClientType

from ansible.module_utils.common.dict_transformations import snake_dict_to_camel_dict

from ansible_collections.amazon.aws.plugins.module_utils.botocore import is_boto3_error_code
from ansible_collections.amazon.aws.plugins.module_utils.modules import AnsibleAWSModule
from ansible_collections.amazon.aws.plugins.module_utils.policy import compare_policies
from ansible_collections.amazon.aws.plugins.module_utils.retries import AWSRetry
from ansible_collections.amazon.aws.plugins.module_utils.s3 import AnsibleS3Error
from ansible_collections.amazon.aws.plugins.module_utils.s3 import AnsibleS3PermissionsError
from ansible_collections.amazon.aws.plugins.module_utils.s3 import AnsibleS3SupportError
from ansible_collections.amazon.aws.plugins.module_utils.s3 import S3ErrorHandler
from ansible_collections.amazon.aws.plugins.module_utils.s3 import delete_s3_bucket
from ansible_collections.amazon.aws.plugins.module_utils.s3 import delete_s3_bucket_accelerate_configuration
from ansible_collections.amazon.aws.plugins.module_utils.s3 import delete_s3_bucket_encryption
from ansible_collections.amazon.aws.plugins.module_utils.s3 import delete_s3_bucket_inventory
from ansible_collections.amazon.aws.plugins.module_utils.s3 import delete_s3_bucket_ownership
from ansible_collections.amazon.aws.plugins.module_utils.s3 import delete_s3_bucket_policy
from ansible_collections.amazon.aws.plugins.module_utils.s3 import delete_s3_bucket_public_access
from ansible_collections.amazon.aws.plugins.module_utils.s3 import delete_s3_bucket_tagging
from ansible_collections.amazon.aws.plugins.module_utils.s3 import get_s3_bucket_accelerate_configuration
from ansible_collections.amazon.aws.plugins.module_utils.s3 import get_s3_bucket_acl
from ansible_collections.amazon.aws.plugins.module_utils.s3 import get_s3_bucket_encryption
from ansible_collections.amazon.aws.plugins.module_utils.s3 import get_s3_bucket_location
from ansible_collections.amazon.aws.plugins.module_utils.s3 import get_s3_bucket_logging
from ansible_collections.amazon.aws.plugins.module_utils.s3 import get_s3_bucket_ownership_controls
from ansible_collections.amazon.aws.plugins.module_utils.s3 import get_s3_bucket_policy
from ansible_collections.amazon.aws.plugins.module_utils.s3 import get_s3_bucket_public_access_block
from ansible_collections.amazon.aws.plugins.module_utils.s3 import get_s3_bucket_request_payment
from ansible_collections.amazon.aws.plugins.module_utils.s3 import get_s3_bucket_tagging
from ansible_collections.amazon.aws.plugins.module_utils.s3 import get_s3_bucket_versioning
from ansible_collections.amazon.aws.plugins.module_utils.s3 import get_s3_object_lock_configuration
from ansible_collections.amazon.aws.plugins.module_utils.s3 import get_s3_waiter
from ansible_collections.amazon.aws.plugins.module_utils.s3 import list_bucket_inventory_configurations
from ansible_collections.amazon.aws.plugins.module_utils.s3 import merge_tags
from ansible_collections.amazon.aws.plugins.module_utils.s3 import normalize_s3_bucket_acls
from ansible_collections.amazon.aws.plugins.module_utils.s3 import normalize_s3_bucket_logging
from ansible_collections.amazon.aws.plugins.module_utils.s3 import normalize_s3_bucket_public_access
from ansible_collections.amazon.aws.plugins.module_utils.s3 import normalize_s3_bucket_versioning
from ansible_collections.amazon.aws.plugins.module_utils.s3 import put_s3_bucket_accelerate_configuration
from ansible_collections.amazon.aws.plugins.module_utils.s3 import put_s3_bucket_acl
from ansible_collections.amazon.aws.plugins.module_utils.s3 import put_s3_bucket_encryption
from ansible_collections.amazon.aws.plugins.module_utils.s3 import put_s3_bucket_inventory
from ansible_collections.amazon.aws.plugins.module_utils.s3 import put_s3_bucket_logging
from ansible_collections.amazon.aws.plugins.module_utils.s3 import put_s3_bucket_ownership
from ansible_collections.amazon.aws.plugins.module_utils.s3 import put_s3_bucket_policy
from ansible_collections.amazon.aws.plugins.module_utils.s3 import put_s3_bucket_public_access
from ansible_collections.amazon.aws.plugins.module_utils.s3 import put_s3_bucket_request_payment
from ansible_collections.amazon.aws.plugins.module_utils.s3 import put_s3_bucket_tagging
from ansible_collections.amazon.aws.plugins.module_utils.s3 import put_s3_bucket_versioning
from ansible_collections.amazon.aws.plugins.module_utils.s3 import put_s3_object_lock_configuration
from ansible_collections.amazon.aws.plugins.module_utils.s3 import s3_acl_to_name
from ansible_collections.amazon.aws.plugins.module_utils.s3 import s3_bucket_exists
from ansible_collections.amazon.aws.plugins.module_utils.s3 import s3_extra_params
from ansible_collections.amazon.aws.plugins.module_utils.s3 import validate_account_regional_bucket_name
from ansible_collections.amazon.aws.plugins.module_utils.s3 import validate_bucket_name
from ansible_collections.amazon.aws.plugins.module_utils.transformation import scrub_none_parameters


def handle_bucket_versioning(s3_client: ClientType, module: AnsibleAWSModule, name: str) -> Tuple[bool, Optional[dict]]:
    """
    Manage versioning for an S3 bucket.
    Parameters:
        s3_client (boto3.client): The Boto3 S3 client object.
        module (AnsibleAWSModule): The Ansible module object.
        name (str): The name of the bucket to handle versioning for.
    Returns:
        A tuple containing a boolean indicating whether versioning
        was changed and a dictionary containing the updated versioning status.
    """

    versioning = module.params.get("versioning")
    required_versioning = ""

    try:
        versioning_status = get_s3_bucket_versioning(s3_client, name)
    except (AnsibleS3PermissionsError, AnsibleS3SupportError) as e:
        if versioning is not None:
            raise
        module.warn(e.message)
        return False, None

    if versioning is None:
        return False, normalize_s3_bucket_versioning(versioning_status)

    if versioning and versioning_status.get("Status") != "Enabled":
        required_versioning = "Enabled"
    if not versioning and versioning_status.get("Status") == "Enabled":
        required_versioning = "Suspended"

    if not required_versioning:
        return False, normalize_s3_bucket_versioning(versioning_status)

    put_s3_bucket_versioning(s3_client, name, required_versioning)

    wait_versioning_is_applied(s3_client, name, required_versioning)
    versioning_status = get_s3_bucket_versioning(s3_client, name)
    return True, normalize_s3_bucket_versioning(versioning_status)


def handle_bucket_requester_pays(
    s3_client: ClientType, module: AnsibleAWSModule, name: str
) -> Tuple[bool, Optional[dict]]:
    """
    Manage requester pays setting for an S3 bucket.
    Parameters:
        s3_client (boto3.client): The Boto3 S3 client object.
        module (AnsibleAWSModule): The Ansible module object.
        name (str): The name of the bucket to handle requester pays setting for.
    Returns:
        A tuple containing a boolean indicating whether requester pays setting
        was changed and a boolean containing the updated requester pays status.
    """

    requester_pays = module.params.get("requester_pays")

    try:
        requester_pays_status = get_s3_bucket_request_payment(s3_client, name)
    except (AnsibleS3PermissionsError, AnsibleS3SupportError) as e:
        if requester_pays is not None:
            raise
        module.warn(e.message)
        return False, None

    requester_pays_status = bool(requester_pays_status == "Requester")

    if requester_pays is None:
        return False, requester_pays_status
    if requester_pays == requester_pays_status:
        return False, requester_pays_status

    payer = "Requester" if requester_pays else "BucketOwner"
    put_s3_bucket_request_payment(s3_client, name, payer)
    requester_pays_status = wait_payer_is_applied(s3_client, name, payer, should_fail=False)
    if requester_pays_status is None:
        put_s3_bucket_request_payment(s3_client, name, payer)
        requester_pays_status = wait_payer_is_applied(s3_client, name, payer, should_fail=True)

    return True, requester_pays


def handle_bucket_public_access_config(
    s3_client: ClientType, module: AnsibleAWSModule, name: str
) -> Tuple[bool, Optional[dict]]:
    """
    Manage public access configuration for an S3 bucket.
    Parameters:
        s3_client (boto3.client): The Boto3 S3 client object.
        module (AnsibleAWSModule): The Ansible module object.
        name (str): The name of the bucket to handle public access configuration for.
    Returns:
        A tuple containing a boolean indicating whether public access configuration
        was changed and a dictionary containing the updated public access configuration.
    """

    public_access = module.params.get("public_access")
    delete_public_access = module.params.get("delete_public_access")

    try:
        current_public_access = get_s3_bucket_public_access_block(s3_client, name)
    except (AnsibleS3PermissionsError, AnsibleS3SupportError) as e:
        if public_access is not None:
            raise
        module.warn(e.message)
        return False, None

    # -- Delete public access block if necessary
    if delete_public_access:
        if not current_public_access:
            return False, None
        delete_s3_bucket_public_access(s3_client, name)
        return True, None

    # -- Create / Update public access block
    # Short circuit comparisons if no change was requested
    if public_access is None:
        return False, normalize_s3_bucket_public_access(current_public_access)

    camel_public_block = snake_dict_to_camel_dict(public_access, capitalize_first=True)
    # No change needed
    if current_public_access == camel_public_block:
        return False, normalize_s3_bucket_public_access(current_public_access)

    # Make a change
    put_s3_bucket_public_access(s3_client, name, public_access_config=camel_public_block)
    return True, normalize_s3_bucket_public_access(camel_public_block)


def handle_bucket_policy(s3_client: ClientType, module: AnsibleAWSModule, name: str) -> Tuple[bool, Optional[dict]]:
    """
    Manage bucket policy for an S3 bucket.
    Parameters:
        s3_client (boto3.client): The Boto3 S3 client object.
        module (AnsibleAWSModule): The Ansible module object.
        name (str): The name of the bucket to handle the policy for.
    Returns:
        A tuple containing a boolean indicating whether the bucket policy
        was changed and a dictionary containing the updated bucket policy.
    """

    policy = module.params.get("policy")
    try:
        current_policy = get_s3_bucket_policy(s3_client, name)
    except (AnsibleS3PermissionsError, AnsibleS3SupportError) as e:
        if policy is not None:
            raise
        module.warn(e.message)
        return False, None

    if policy is None:
        return False, current_policy

    if isinstance(policy, str):
        try:
            policy = json.loads(policy)
        except ValueError as e:
            raise AnsibleS3Error(exception=e, message="Unable to parse bucket policy") from e

    if not policy:
        if not current_policy:
            return False, current_policy

        delete_s3_bucket_policy(s3_client, name)
        return True, wait_policy_is_applied(s3_client, name, policy)

    if not compare_policies(current_policy, policy):
        return False, current_policy

    put_s3_bucket_policy(s3_client, name, policy)
    current_policy = wait_policy_is_applied(s3_client, name, policy, should_fail=False)
    if current_policy is None:
        # As for request payement, it happens quite a lot of times that the put request was not taken into
        # account, so we retry one more time
        put_s3_bucket_policy(s3_client, name, policy)
        current_policy = wait_policy_is_applied(s3_client, name, policy, should_fail=True)

    return True, current_policy


def handle_bucket_tags(s3_client: ClientType, module: AnsibleAWSModule, name: str) -> Tuple[bool, Optional[dict]]:
    """
    Manage tags for an S3 bucket.
    Parameters:
        s3_client (boto3.client): The Boto3 S3 client object.
        module (AnsibleAWSModule): The Ansible module object.
        name (str): The name of the bucket to handle tags for.
    Returns:
        A tuple containing a boolean indicating whether tags were changed
        and a dictionary containing the updated tags.
    """

    tags = module.params.get("tags")
    purge_tags = cast(bool, module.params.get("purge_tags"))

    try:
        current_tags_dict = get_s3_bucket_tagging(s3_client, name)
    except (AnsibleS3PermissionsError, AnsibleS3SupportError) as e:
        if tags is not None:
            raise
        module.warn(e.message)
        return False, None

    if tags is None:
        return False, current_tags_dict

    updated_tags = merge_tags(current_tags_dict, tags, purge_tags)

    if current_tags_dict == updated_tags:
        return False, current_tags_dict

    if not updated_tags:
        delete_s3_bucket_tagging(s3_client, name)
    else:
        put_s3_bucket_tagging(s3_client, name, updated_tags)
    current_tags_dict = wait_tags_are_applied(s3_client, name, updated_tags)

    return True, updated_tags


def _handle_remove_bucket_encryption(
    s3_client: ClientType, name: str, current_encryption: Optional[dict]
) -> Tuple[bool, Optional[dict]]:
    current_encryption_algorithm = current_encryption.get("SSEAlgorithm") if current_encryption else None

    if current_encryption_algorithm is None:
        return False, current_encryption

    delete_s3_bucket_encryption(s3_client, name)
    current_encryption = wait_encryption_is_applied(s3_client, name, None)

    return True, current_encryption


def _handle_bucket_encryption(
    s3_client: ClientType,
    name: str,
    current_encryption: Optional[dict],
    encryption: Optional[str],
    encryption_key_id: Optional[str],
) -> Tuple[bool, Optional[dict]]:
    if encryption is None:
        return False, current_encryption

    current_encryption_algorithm = cast(dict, current_encryption).get("SSEAlgorithm") if current_encryption else None
    current_encryption_key = cast(dict, current_encryption).get("KMSMasterKeyID") if current_encryption else None

    if encryption == current_encryption_algorithm:
        # Non KMS is simple
        if encryption != "aws:kms":
            return False, current_encryption
        # When working with KMS, we also need to check the Key
        if current_encryption_key == encryption_key_id:
            return False, current_encryption

    expected_encryption = {"SSEAlgorithm": encryption}
    if encryption_key_id is not None:
        expected_encryption["KMSMasterKeyID"] = encryption_key_id

    current_encryption = put_bucket_encryption_with_retry(s3_client, name, expected_encryption)
    return True, current_encryption


def _handle_bucket_key_encryption(
    s3_client: ClientType, name, current_encryption, bucket_key_enabled
) -> Tuple[bool, Optional[dict]]:
    if bucket_key_enabled is None:
        return False, current_encryption

    current_encryption_algorithm = current_encryption.get("SSEAlgorithm") if current_encryption else None

    if current_encryption_algorithm != "aws:kms":
        if bucket_key_enabled:
            raise AnsibleS3Error(
                f'Unable to set bucket key: current encryption algorithm ("{current_encryption_algorithm}") is not "aws:kms"'
            )
        return False, current_encryption

    if get_bucket_key_enabled(s3_client, name) == bucket_key_enabled:
        return False, current_encryption

    expected_encryption = bool(bucket_key_enabled)
    put_bucket_key_with_retry(s3_client, name, expected_encryption)
    return True, get_bucket_encryption(s3_client, name)


def handle_bucket_encryption(s3_client: ClientType, module: AnsibleAWSModule, name: str) -> Tuple[bool, Optional[dict]]:
    """
    Manage encryption settings for an S3 bucket.
    Parameters:
        s3_client (boto3.client): The Boto3 S3 client object.
        module (AnsibleAWSModule): The Ansible module object.
        name (str): The name of the bucket to handle encryption for.
    Returns:
        A tuple containing a boolean indicating whether encryption settings
        were changed and a dictionary containing the updated encryption settings.
    """
    encryption = module.params.get("encryption")
    encryption_key_id = module.params.get("encryption_key_id")
    bucket_key_enabled = module.params.get("bucket_key_enabled")

    try:
        current_encryption = get_bucket_encryption(s3_client, name)
    except (AnsibleS3PermissionsError, AnsibleS3SupportError) as e:
        if encryption is not None:
            raise
        module.warn(e.message)
        return False, None

    if encryption == "none":
        return _handle_remove_bucket_encryption(s3_client, name, current_encryption)

    changed, current_encryption = _handle_bucket_encryption(
        s3_client, name, current_encryption, encryption, encryption_key_id
    )
    bk_changed, current_encryption = _handle_bucket_key_encryption(
        s3_client, name, current_encryption, bucket_key_enabled
    )
    changed |= bk_changed
    return changed, current_encryption


def handle_bucket_ownership(s3_client: ClientType, module: AnsibleAWSModule, name: str) -> Tuple[bool, Optional[str]]:
    """
    Manage ownership settings for an S3 bucket.
    Parameters:
        s3_client (boto3.client): The Boto3 S3 client object.
        module (AnsibleAWSModule): The Ansible module object.
        name (str): The name of the bucket to handle ownership for.
    Returns:
        A tuple containing a boolean indicating whether ownership settings were changed
        and a dictionary containing the updated ownership settings.
    """

    delete_object_ownership = module.params.get("delete_object_ownership")
    object_ownership = module.params.get("object_ownership")

    try:
        bucket_ownership = get_bucket_ownership_cntrl(s3_client, name)
    except (AnsibleS3PermissionsError, AnsibleS3SupportError) as e:
        if delete_object_ownership or object_ownership is not None:
            raise
        module.warn(e.message)
        return False, None

    if delete_object_ownership:
        if bucket_ownership is None:
            return False, None
        delete_s3_bucket_ownership(s3_client, name)
        return True, None

    if object_ownership is None:
        return False, bucket_ownership
    if bucket_ownership == object_ownership:
        return False, bucket_ownership

    put_s3_bucket_ownership(s3_client, name, object_ownership=object_ownership)
    return True, object_ownership


def handle_bucket_acl(
    s3_client: ClientType, module: AnsibleAWSModule, name: str
) -> Tuple[bool, Optional[str], Optional[dict]]:
    """
    Manage Access Control List (ACL) for an S3 bucket.
    Parameters:
        s3_client (boto3.client): The Boto3 S3 client object.
        module (AnsibleAWSModule): The Ansible module object.
        name (str): The name of the bucket to handle ACL for.
    Returns:
        A tuple containing a boolean indicating whether ACL was changed and a dictionary containing the updated ACL.
    """

    acl = module.params.get("acl")

    try:
        bucket_acl = get_s3_bucket_acl(s3_client, name)
        current_name = s3_acl_to_name(bucket_acl)
    except (AnsibleS3PermissionsError, AnsibleS3SupportError) as e:
        if acl is not None:
            raise
        module.warn(e.message)
        return False, None, None

    if acl is None or acl == current_name:
        return False, current_name, bucket_acl

    put_s3_bucket_acl(s3_client, name, acl)
    new_acl = get_s3_bucket_acl(s3_client, name)
    new_name = s3_acl_to_name(new_acl)
    return True, new_name, new_acl


def handle_bucket_object_lock(s3_client: ClientType, module: AnsibleAWSModule, name: str) -> bool:
    """
    Manage object lock configuration for an S3 bucket.
    Parameters:
        s3_client (boto3.client): The Boto3 S3 client object.
        module (AnsibleAWSModule): The Ansible module object.
        name (str): The name of the bucket to handle object lock for.
    Returns:
        The updated object lock configuration.
    """

    object_lock_enabled = module.params.get("object_lock_enabled")

    try:
        object_lock_status = get_bucket_object_lock_enabled(s3_client, name)
    except (AnsibleS3PermissionsError, AnsibleS3SupportError) as e:
        if object_lock_enabled is not None:
            raise
        module.warn(e.message)
        return False

    if object_lock_enabled is None:
        return object_lock_status

    if not object_lock_enabled and object_lock_status:
        raise AnsibleS3Error("Disabling object lock for existing buckets is not supported")
    if object_lock_enabled and not object_lock_status:
        raise AnsibleS3Error("Enabling object lock for existing buckets is not supported")

    return object_lock_status


def handle_bucket_accelerate(s3_client: ClientType, module: AnsibleAWSModule, name: str) -> Tuple[bool, Optional[bool]]:
    """
    Manage transfer accelerate for an S3 bucket.
    Parameters:
        s3_client (boto3.client): The Boto3 S3 client object.
        module (AnsibleAWSModule): The Ansible module object.
        name (str): The name of the bucket to handle transfer accelerate for.
    Returns:
        A tuple containing a boolean indicating whether transfer accelerate setting was changed
        and a boolean indicating the transfer accelerate status.
    """

    accelerate_enabled = module.params.get("accelerate_enabled")

    try:
        accelerate_status = get_bucket_accelerate_status(s3_client, name)
    except (AnsibleS3PermissionsError, AnsibleS3SupportError) as e:
        if accelerate_enabled is not None:
            raise
        module.warn(e.message)
        return False, None

    if accelerate_enabled is None:
        return False, accelerate_status
    if accelerate_enabled == accelerate_status:
        return False, accelerate_status

    if not accelerate_enabled:
        delete_s3_bucket_accelerate_configuration(s3_client, name)
        return True, False

    put_s3_bucket_accelerate_configuration(s3_client, name)

    return True, True


def handle_bucket_object_lock_retention(
    s3_client: ClientType, module: AnsibleAWSModule, name: str
) -> Tuple[bool, Optional[dict]]:
    """
    Manage object lock retention configuration for an S3 bucket.
    Parameters:
        s3_client (boto3.client): The Boto3 S3 client object.
        module (AnsibleAWSModule): The Ansible module object.
        name (str): The name of the bucket to handle object lock for.
    Returns:
        A tuple containing a boolean indicating whether the bucket object lock
        retention configuration was changed and a dictionary containing the change.
    """

    object_lock_enabled = module.params.get("object_lock_enabled")
    object_lock_default_retention = module.params.get("object_lock_default_retention")

    try:
        if object_lock_enabled:
            object_lock_configuration_status = get_object_lock_configuration(s3_client, name)
        else:
            object_lock_configuration_status = {}
    except (AnsibleS3PermissionsError, AnsibleS3SupportError) as e:
        if object_lock_default_retention is not None:
            raise
        module.warn(e.message)
        return False, None

    if object_lock_default_retention is None:
        return False, object_lock_configuration_status

    if not object_lock_default_retention and object_lock_configuration_status:
        raise AnsibleS3Error("Removing object lock default retention is not supported")

    conf = scrub_none_parameters(snake_dict_to_camel_dict(object_lock_default_retention, capitalize_first=True))
    if object_lock_configuration_status == conf:
        return False, object_lock_configuration_status

    put_s3_object_lock_configuration(s3_client, name, conf)
    return True, object_lock_default_retention


def handle_bucket_inventory(s3_client: ClientType, module: AnsibleAWSModule, name: str) -> Tuple[bool, Optional[list]]:
    """
    Manage inventory configuration for an S3 bucket.
    Parameters:
        s3_client (boto3.client): The Boto3 S3 client object.
        module (AnsibleAWSModule): The Ansible module object.
        name (str): The name of the bucket to handle inventory for.
    Returns:
        A tuple containing a boolean indicating whether inventory settings were changed
        and a dictionary containing the updated inventory.
    """
    declared_inventories = module.params.get("inventory")

    try:
        current_inventories = list_bucket_inventory_configurations(s3_client, name)
        present_inventories = {i["Id"]: i for i in current_inventories}
    except (AnsibleS3PermissionsError, AnsibleS3SupportError) as e:
        if declared_inventories is not None:
            raise
        module.warn(e.message)
        return False, None

    if declared_inventories is None:
        return (False, current_inventories)

    declared_inventories = cast(list, declared_inventories)
    bucket_changed = False
    results = []

    for declared_inventory in declared_inventories:
        camel_destination = snake_dict_to_camel_dict(declared_inventory.get("destination", {}), True)
        declared_inventory_api = {
            "IsEnabled": True,
            "Id": declared_inventory.get("id"),
            "Destination": {"S3BucketDestination": {k: v for k, v in camel_destination.items() if v is not None}},
            "IncludedObjectVersions": declared_inventory.get("included_object_versions"),
            "Schedule": {"Frequency": declared_inventory.get("schedule")},
            "OptionalFields": [],
        }
        for field in declared_inventory.get("optional_fields", []):
            declared_inventory_api["OptionalFields"].append(field)
        if declared_inventory.get("filter") is not None:
            declared_inventory_api["Filter"] = {"Prefix": declared_inventory.get("filter")}

        present_inventory = present_inventories.pop(declared_inventory_api["Id"], None)

        if declared_inventory_api != present_inventory:
            put_s3_bucket_inventory(s3_client, name, declared_inventory_api)
            bucket_changed = True

        results.append(declared_inventory_api)

    for inventory_id in present_inventories.keys():
        delete_s3_bucket_inventory(s3_client, name, inventory_id)
        bucket_changed = True

    return bucket_changed, results


def handle_bucket_logging(s3_client: ClientType, module: AnsibleAWSModule, name: str) -> Tuple[bool, Optional[dict]]:
    """
    Manage logging configuration for an S3 bucket.
    Parameters:
        s3_client (boto3.client): The Boto3 S3 client object.
        module (AnsibleAWSModule): The Ansible module object.
        name (str): The name of the bucket to handle logging for.
    Returns:
        A tuple containing a boolean indicating whether logging settings were changed
        and a dictionary containing the updated logging configuration.
    """
    logging_param = module.params.get("logging")

    try:
        current_logging = get_s3_bucket_logging(s3_client, name)
    except (AnsibleS3PermissionsError, AnsibleS3SupportError) as e:
        if logging_param is not None:
            raise
        module.warn(e.message)
        return False, None

    # Normalize current logging state
    current_normalized = normalize_s3_bucket_logging(current_logging)

    # If no logging parameter specified, return current state
    if logging_param is None:
        return False, current_normalized

    # Determine desired state
    # Disable logging if empty dict or target_bucket is empty string
    if logging_param == {} or logging_param.get("target_bucket") == "":
        desired_config = {}
        desired_normalized = None
    else:
        # Enable logging with specified configuration
        target_bucket = logging_param.get("target_bucket")
        target_prefix = logging_param.get("target_prefix", "")

        desired_config = {
            "LoggingEnabled": {
                "TargetBucket": target_bucket,
                "TargetPrefix": target_prefix,
            }
        }
        desired_normalized = {
            "target_bucket": target_bucket,
            "target_prefix": target_prefix,
        }

    # Compare current and desired states
    if current_normalized == desired_normalized:
        return False, current_normalized

    # Make the change
    put_s3_bucket_logging(s3_client, name, desired_config)

    return True, desired_normalized


def create_or_update_bucket(s3_client: ClientType, module: AnsibleAWSModule) -> NoReturn:
    """
    Create or update an S3 bucket along with its associated configurations.
    This function creates a new S3 bucket if it does not already exist, and updates its configurations,
    such as versioning, requester pays, public access block configuration, policy, tags, encryption, bucket ownership,
    ACL, and object lock settings. It returns whether any changes were made and the updated configurations.
    Parameters:
        s3_client (boto3.client): The Boto3 S3 client object.
        module (AnsibleAWSModule): The Ansible module object.
    Returns:
        None
    """
    name = cast(str, module.params.get("name"))
    object_lock_enabled = module.params.get("object_lock_enabled")
    location = get_s3_bucket_location(module)

    changed = False
    result: dict = {}

    bucket_is_present = s3_bucket_exists(s3_client, name)

    if not bucket_is_present:
        account_regional = module.params.get("account_regional")
        changed = create_bucket(s3_client, name, location, object_lock_enabled, account_regional)
        waiter = get_s3_waiter(s3_client, "bucket_exists")
        S3ErrorHandler.common_error_handler(f"wait for bucket {name} to be created")(waiter.wait)(Bucket=name)

    # Versioning
    versioning_changed, versioning_result = handle_bucket_versioning(s3_client, module, name)
    result["versioning"] = versioning_result

    # Requester pays
    requester_pays_changed, requester_pays_result = handle_bucket_requester_pays(s3_client, module, name)
    result["requester_pays"] = requester_pays_result

    # Public access clock configuration
    public_access_config_changed, public_access_config_result = handle_bucket_public_access_config(
        s3_client, module, name
    )
    result["public_access_block"] = public_access_config_result

    # Policy
    policy_changed, current_policy = handle_bucket_policy(s3_client, module, name)
    result["policy"] = current_policy

    # Tags
    tags_changed, current_tags_dict = handle_bucket_tags(s3_client, module, name)
    result["tags"] = current_tags_dict

    # Encryption
    encryption_changed, current_encryption = handle_bucket_encryption(s3_client, module, name)
    result["encryption"] = current_encryption

    # -- Bucket ownership
    bucket_ownership_changed, object_ownership_result = handle_bucket_ownership(s3_client, module, name)
    result["object_ownership"] = object_ownership_result

    # -- Bucket ACL
    bucket_acl_changed, bucket_acl_result, acl_info = handle_bucket_acl(s3_client, module, name)
    result["acl"] = bucket_acl_result
    result["acl_grants"] = normalize_s3_bucket_acls(acl_info)

    # -- Transfer Acceleration
    bucket_accelerate_changed, bucket_accelerate_result = handle_bucket_accelerate(s3_client, module, name)
    result["accelerate_enabled"] = bucket_accelerate_result

    # TODO - MERGE handle_bucket_object_lock and handle_bucket_object_lock_retention
    # -- Object Lock
    bucket_object_lock_result = handle_bucket_object_lock(s3_client, module, name)
    result["object_lock_enabled"] = bucket_object_lock_result

    # -- Object Lock Default Retention
    bucket_object_lock_retention_changed, bucket_object_lock_retention_result = handle_bucket_object_lock_retention(
        s3_client, module, name
    )
    result["object_lock_default_retention"] = bucket_object_lock_retention_result
    # -- Inventory
    bucket_inventory_changed, bucket_inventory_result = handle_bucket_inventory(s3_client, module, name)
    result["bucket_inventory"] = bucket_inventory_result

    # -- Logging
    bucket_logging_changed, bucket_logging_result = handle_bucket_logging(s3_client, module, name)
    result["logging"] = bucket_logging_result

    # Module exit
    changed = (
        changed
        or versioning_changed
        or requester_pays_changed
        or public_access_config_changed
        or policy_changed
        or tags_changed
        or encryption_changed
        or bucket_ownership_changed
        or bucket_acl_changed
        or bucket_accelerate_changed
        or bucket_object_lock_retention_changed
        or bucket_inventory_changed
        or bucket_logging_changed
    )
    module.exit_json(changed=changed, name=name, **result)


@S3ErrorHandler.common_error_handler("create S3 bucket")
@AWSRetry.jittered_backoff(max_delay=120, catch_extra_error_codes=["OperationAborted"])
def _create_bucket(s3_client: ClientType, **params) -> bool:
    try:
        s3_client.create_bucket(**params)
        return True
    except is_boto3_error_code("BucketAlreadyOwnedByYou"):
        # We should never get here since we check the bucket presence before calling the create_or_update_bucket
        # method. However, the AWS Api sometimes fails to report bucket presence, so we catch this exception
        return False


def create_bucket(
    s3_client: ClientType,
    bucket_name: str,
    location: str,
    object_lock_enabled: Optional[bool] = False,
    account_regional: Optional[bool] = None,
) -> bool:
    """
    Create an S3 bucket.
    Parameters:
        s3_client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the bucket to create.
        location (str): The AWS region where the bucket should be created. If None, it defaults to "us-east-1".
        object_lock_enabled (bool): Whether to enable object lock for the bucket. Defaults to False.
        account_regional (bool): Whether to create the bucket in the account-regional namespace. Defaults to None (global namespace).
    Returns:
        True if the bucket was successfully created, False otherwise.
    """
    params: dict = {"Bucket": bucket_name}

    configuration: dict = {}
    if location not in ("us-east-1", None):
        configuration["LocationConstraint"] = location

    if configuration:
        params["CreateBucketConfiguration"] = configuration

    if object_lock_enabled is not None:
        params["ObjectLockEnabledForBucket"] = object_lock_enabled
    if account_regional:
        params["BucketNamespace"] = "account-regional"
    return _create_bucket(s3_client, **params)


def get_object_lock_configuration(s3_client: ClientType, bucket_name: str) -> Optional[dict]:
    """
    Get the object lock default retention configuration for an S3 bucket.
    Parameters:
        s3_client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the S3 bucket.
    Returns:
        Object lock default retention configuration dictionary.
    """
    result = get_s3_object_lock_configuration(s3_client, bucket_name)
    return result.get("Rule", {}).get("DefaultRetention", {})


def get_bucket_accelerate_status(s3_client: ClientType, bucket_name: str) -> bool:
    """
    Get transfer accelerate status of the S3 bucket.
    Parameters:
        s3_client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the S3 bucket.
    Returns:
        Transfer accelerate status of the S3 bucket.
    """
    accelerate_configuration = get_s3_bucket_accelerate_configuration(s3_client, bucket_name)
    return accelerate_configuration.get("Status") == "Enabled"


def get_bucket_object_lock_enabled(s3_client: ClientType, bucket_name: str) -> bool:
    """
    Retrieve the object lock configuration status for an S3 bucket.
    Parameters:
        s3_client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the S3 bucket.
    Returns:
        True if object lock is enabled for the bucket, False otherwise.
    """
    object_lock_configuration = get_s3_object_lock_configuration(s3_client, bucket_name)
    return object_lock_configuration.get("ObjectLockEnabled") == "Enabled"


def get_bucket_encryption(s3_client: ClientType, bucket_name: str) -> Optional[dict]:
    """
    Retrieve the encryption configuration for an S3 bucket.
    Parameters:
        s3_client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the S3 bucket.
    Returns:
        Encryption configuration of the bucket.
    """
    try:
        result = get_s3_bucket_encryption(s3_client, bucket_name)
        return result.get("Rules", [])[0].get("ApplyServerSideEncryptionByDefault")
    except (IndexError, KeyError):
        return None


def get_bucket_key_enabled(s3_client: ClientType, bucket_name: str) -> Optional[bool]:
    """
    Retrieve the status of server-side encryption for an S3 bucket.
    Parameters:
        s3_client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the S3 bucket.
    Returns:
        Whether or not if server-side encryption is enabled for the bucket.
    """
    try:
        result = get_s3_bucket_encryption(s3_client, bucket_name)
        return result.get("Rules", [])[0].get("BucketKeyEnabled")
    except (IndexError, KeyError):
        return None


def put_bucket_encryption_with_retry(s3_client: ClientType, name: str, expected_encryption: dict) -> dict:
    """
    Set the encryption configuration for an S3 bucket with retry logic.
    Parameters:
        s3_client (boto3.client): The Boto3 S3 client object.
        name (str): The name of the S3 bucket.
        expected_encryption (dict): A dictionary containing the expected encryption configuration.
    Returns:
        Updated encryption configuration of the bucket.
    """
    max_retries = 3
    for retries in range(1, max_retries + 1):
        put_s3_bucket_encryption(s3_client, name, expected_encryption)
        current_encryption = wait_encryption_is_applied(
            s3_client, name, expected_encryption, should_fail=(retries == max_retries), retries=5
        )
        if current_encryption == expected_encryption:
            return current_encryption

    # We shouldn't get here, the only time this should happen is if
    # current_encryption != expected_encryption and retries == max_retries
    raise AnsibleS3Error("Failed to set bucket encryption configuration")


def put_bucket_key_with_retry(s3_client: ClientType, name: str, expected_encryption: bool) -> None:
    """
    Set the status of server-side encryption for an S3 bucket.
    Parameters:
        module (AnsibleAWSModule): The Ansible module object.
        s3_client (boto3.client): The Boto3 S3 client object.
        name (str): The name of the S3 bucket.
        expected_encryption (bool): The expected status of server-side encryption using AWS KMS.
    Returns:
        The updated status of server-side encryption using AWS KMS for the bucket.
    """
    max_retries = 3
    for retries in range(1, max_retries + 1):
        put_bucket_key(s3_client, name, expected_encryption)
        try:
            wait_bucket_key_is_applied(s3_client, name, expected_encryption, retries=5)
            return
        except AnsibleS3Error:
            if retries == max_retries:
                raise
            pass

    # We shouldn't get here, the only time this should happen is if
    # current_encryption != expected_encryption and retries == max_retries
    raise AnsibleS3Error("Failed to set bucket key")


@S3ErrorHandler.common_error_handler("set bucket key based encryption")
@AWSRetry.jittered_backoff(max_delay=120, catch_extra_error_codes=["NoSuchBucket", "OperationAborted"])
def put_bucket_key(s3_client: ClientType, bucket_name: str, encryption: bool) -> None:
    """
    Set the status of server-side encryption for an S3 bucket.
    Parameters:
        s3_client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the S3 bucket.
        encryption (bool): The status of server-side encryption using AWS KMS.
    Returns:
        None
    """
    # server_side_encryption_configuration ={'Rules': [{'BucketKeyEnabled': encryption}]}
    encryption_status = get_s3_bucket_encryption(s3_client, bucket_name)
    encryption_status["Rules"][0]["BucketKeyEnabled"] = encryption
    s3_client.put_bucket_encryption(Bucket=bucket_name, ServerSideEncryptionConfiguration=encryption_status)


def wait_policy_is_applied(
    s3_client: ClientType, bucket_name: str, expected_policy: dict, should_fail: bool = True
) -> Optional[dict]:
    """
    Wait for a bucket policy to be applied to an S3 bucket.
    Parameters:
        module (AnsibleAWSModule): The Ansible module object.
        s3_client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the S3 bucket.
        expected_policy (dict): The expected bucket policy.
        should_fail (bool): Flag indicating whether to fail if the policy is not applied within the expected time. Default is True.
    Returns:
        The current policy applied to the bucket, or None if the policy failed to apply within the expected time.
    """
    for dummy in range(0, 12):
        current_policy = get_s3_bucket_policy(s3_client, bucket_name)
        if not compare_policies(current_policy, expected_policy):
            return current_policy
        time.sleep(5)

    if should_fail:
        raise AnsibleS3Error(message="Bucket policy failed to apply in the expected time")
    return None


def wait_payer_is_applied(
    s3_client: ClientType, bucket_name: str, expected_payer: str, should_fail=True
) -> Optional[str]:
    """
    Wait for the requester pays setting to be applied to an S3 bucket.
    Parameters:
        s3_client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the S3 bucket.
        expected_payer (bool): The expected status of the requester pays setting.
        should_fail (bool): Flag indicating whether to fail if the setting is not applied within the expected time. Default is True.
    Returns:
        The current status of the requester pays setting applied to the bucket.
    """
    for dummy in range(0, 12):
        requester_pays_status = get_s3_bucket_request_payment(s3_client, bucket_name)
        if requester_pays_status == expected_payer:
            return requester_pays_status
        time.sleep(5)

    if should_fail:
        raise AnsibleS3Error(message="Bucket request payer setting failed to apply in the expected time")
    return None


def wait_encryption_is_applied(
    s3_client: ClientType,
    bucket_name: str,
    expected_encryption: Optional[dict],
    should_fail: bool = True,
    retries: int = 12,
) -> Optional[dict]:
    """
    Wait for the encryption setting to be applied to an S3 bucket.
    Parameters:
        module (AnsibleAWSModule): The Ansible module object.
        s3_client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the S3 bucket.
        expected_encryption(dict): The expected encryption setting.
        should_fail (bool): Flag indicating whether to fail if the setting is not applied within the expected time. Default is True.
        retries (int): The number of retries to attempt. Default is 12.
    Returns:
        The current encryption setting applied to the bucket.
    """
    for dummy in range(0, retries):
        encryption = get_bucket_encryption(s3_client, bucket_name)
        if encryption != expected_encryption:
            time.sleep(5)
        else:
            return encryption

    if should_fail:
        raise AnsibleS3Error("Bucket encryption failed to apply in the expected time")

    return encryption


def wait_bucket_key_is_applied(
    s3_client: ClientType, bucket_name: str, expected_encryption: bool, retries: int = 12
) -> None:
    """
    Wait for the bucket key setting to be applied to an S3 bucket.
    Parameters:
        module (AnsibleAWSModule): The Ansible module object.
        s3_client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the S3 bucket.
        expected_encryption (bool): The expected bucket key setting.
        should_fail (bool): Flag indicating whether to fail if the setting is not applied within the expected time. Default is True.
        retries (int): The number of retries to attempt. Default is 12.
    Returns:
        The current bucket key setting applied to the bucket.
    """
    extension = "enabled" if expected_encryption else "disabled"
    waiter_name = f"bucket_key_encryption_{extension}"
    waiter = get_s3_waiter(s3_client, waiter_name)
    S3ErrorHandler.common_error_handler(f"wait for bucket key encryption to be {extension}")(waiter.wait)(
        Bucket=bucket_name
    )


def wait_versioning_is_applied(s3_client: ClientType, bucket_name: str, required_versioning: str) -> None:
    """
    Wait for the versioning setting to be applied to an S3 bucket.
    Parameters:
        s3_client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the S3 bucket.
        required_versioning (dict): The required versioning status.
    Returns:
        The current versioning status applied to the bucket.
    """
    waiter_name = "bucket_versioning_enabled" if required_versioning == "Enabled" else "bucket_versioning_suspended"
    waiter = get_s3_waiter(s3_client, waiter_name)
    S3ErrorHandler.common_error_handler(f"wait for bucket versioning to be {required_versioning}")(waiter.wait)(
        Bucket=bucket_name
    )


def wait_tags_are_applied(s3_client: ClientType, bucket_name: str, expected_tags_dict: dict) -> dict:
    """
    Wait for the tags to be applied to an S3 bucket.
    Parameters:
        s3_client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the S3 bucket.
        expected_tags_dict (dict): The expected tags dictionary.
    Returns:
        The current tags dictionary applied to the bucket.
    """
    for dummy in range(0, 12):
        current_tags_dict = get_s3_bucket_tagging(s3_client, bucket_name)
        if current_tags_dict == expected_tags_dict:
            return current_tags_dict
        time.sleep(5)

    raise AnsibleS3Error(message="Bucket tags failed to apply in the expected time")


def get_bucket_ownership_cntrl(s3_client, bucket_name: str) -> Optional[str]:
    """
    Get the current bucket ownership controls.
    Parameters:
        s3_client (boto3.client): The Boto3 S3 client object.
        bucket_name (str): The name of the S3 bucket.
    Returns:
      The object ownership rule
    """
    result = get_s3_bucket_ownership_controls(s3_client, bucket_name)
    if not result:
        return None
    try:
        return result["Rules"][0]["ObjectOwnership"]
    except (KeyError, IndexError) as e:
        raise AnsibleS3SupportError(message="Failed to parse bucket object ownership settings") from e


def paginated_list(s3_client: ClientType, **pagination_params) -> Iterator[List[str]]:
    """
    Paginate through the list of objects in an S3 bucket.
    This function yields the keys of objects in the S3 bucket, paginating through the results.
    Parameters:
        s3_client (boto3.client): The Boto3 S3 client object.
        **pagination_params: Additional parameters to pass to the paginator.
    Yields:
        list: A list of keys of objects in the bucket for each page of results.
    """
    pg = s3_client.get_paginator("list_objects_v2")
    for page in pg.paginate(**pagination_params):
        yield [data["Key"] for data in page.get("Contents", [])]


def paginated_versions_list(s3_client: ClientType, **pagination_params) -> Iterator[List[Tuple[str, str]]]:
    """
    Paginate through the list of object versions in an S3 bucket.
    This function yields the keys and version IDs of object versions in the S3 bucket, paginating through the results.
    Parameters:
        s3_client (boto3.client): The Boto3 S3 client object.
        **pagination_params: Additional parameters to pass to the paginator.
    Yields:
        list: A list of tuples containing keys and version IDs of object versions in the bucket for each page of results.
    """
    try:
        pg = s3_client.get_paginator("list_object_versions")
        for page in pg.paginate(**pagination_params):
            # We have to merge the Versions and DeleteMarker lists here, as DeleteMarkers can still prevent a bucket deletion
            yield [
                (data["Key"], data["VersionId"]) for data in (page.get("Versions", []) + page.get("DeleteMarkers", []))
            ]
    except is_boto3_error_code("NoSuchBucket"):
        yield []


@S3ErrorHandler.deletion_error_handler("delete objects in bucket")
@AWSRetry.jittered_backoff(max_delay=120, catch_extra_error_codes=["OperationAborted"])
def delete_objects(s3_client: ClientType, module: AnsibleAWSModule, name: str) -> None:
    """
    Delete objects from an S3 bucket.
    Parameters:
        s3_client (boto3.client): The Boto3 S3 client object.
        module (AnsibleAWSModule): The Ansible module object.
        name (str): The name of the S3 bucket.
    Returns:
        None
    """
    for key_version_pairs in paginated_versions_list(s3_client, Bucket=name):
        formatted_keys = [{"Key": key, "VersionId": version} for key, version in key_version_pairs]
        for fk in formatted_keys:
            # remove VersionId from cases where they are `None` so that
            # unversioned objects are deleted using `DeleteObject`
            # rather than `DeleteObjectVersion`, improving backwards
            # compatibility with older IAM policies.
            if not fk.get("VersionId") or fk.get("VersionId") == "null":
                fk.pop("VersionId")
        if formatted_keys:
            resp = s3_client.delete_objects(Bucket=name, Delete={"Objects": formatted_keys})
            if resp.get("Errors"):
                objects_to_delete = ", ".join([k["Key"] for k in resp["Errors"]])
                raise AnsibleS3Error(
                    f"Could not empty bucket before deleting. Could not delete objects: {objects_to_delete}",
                    errors=resp["Errors"],
                )


def destroy_bucket(s3_client: ClientType, module: AnsibleAWSModule) -> None:
    """
    This function destroys an S3 bucket.
    Parameters:
        s3_client (boto3.client): The Boto3 S3 client object.
        module (AnsibleAWSModule): The Ansible module object.
    Returns:
        None
    """
    force = module.params.get("force")
    name = cast(str, module.params.get("name"))
    bucket_is_present = s3_bucket_exists(s3_client, name)

    if not bucket_is_present:
        module.exit_json(changed=False)

    if force:
        # if there are contents then we need to delete them (including versions) before we can delete the bucket
        delete_objects(s3_client, module, name)

    delete_s3_bucket(s3_client, name)
    waiter = get_s3_waiter(s3_client, "bucket_not_exists")
    S3ErrorHandler.deletion_error_handler(f"wait for bucket f{name} to be deleted")(waiter.wait)(Bucket=name)

    module.exit_json(changed=True)


def main():
    argument_spec = dict(
        name=dict(required=True),
        validate_bucket_name=dict(type="bool", default=True),
        dualstack=dict(default=False, type="bool"),
        state=dict(default="present", choices=["present", "absent"]),
        ceph=dict(default=False, type="bool", aliases=["rgw"]),
        account_regional=dict(type="bool"),
        # ** Warning **
        # we support non-AWS implementations, only force/purge options should have a
        # default set for any top-level option.  We need to be able to identify
        # unset options where we can ignore NotImplemented exceptions.
        tags=dict(type="dict", aliases=["resource_tags"]),
        purge_tags=dict(type="bool", default=True),
        force=dict(default=False, type="bool"),
        policy=dict(type="json"),
        versioning=dict(type="bool"),
        requester_pays=dict(type="bool"),
        encryption=dict(choices=["none", "AES256", "aws:kms"]),
        encryption_key_id=dict(),
        bucket_key_enabled=dict(type="bool"),
        public_access=dict(
            type="dict",
            options=dict(
                block_public_acls=dict(type="bool", default=False),
                ignore_public_acls=dict(type="bool", default=False),
                block_public_policy=dict(type="bool", default=False),
                restrict_public_buckets=dict(type="bool", default=False),
            ),
        ),
        delete_public_access=dict(type="bool", default=False),
        object_ownership=dict(type="str", choices=["BucketOwnerEnforced", "BucketOwnerPreferred", "ObjectWriter"]),
        delete_object_ownership=dict(type="bool", default=False),
        acl=dict(type="str", choices=["private", "public-read", "public-read-write", "authenticated-read"]),
        accelerate_enabled=dict(type="bool"),
        object_lock_enabled=dict(type="bool"),
        object_lock_default_retention=dict(
            type="dict",
            options=dict(
                mode=dict(type="str", choices=["GOVERNANCE", "COMPLIANCE"], required=True),
                years=dict(type="int"),
                days=dict(type="int"),
            ),
            mutually_exclusive=[("days", "years")],
            required_one_of=[("days", "years")],
        ),
        inventory=dict(
            type="list",
            elements="dict",
            options=dict(
                destination=dict(
                    type="dict",
                    options=dict(
                        account_id=dict(type="str"),
                        bucket=dict(type="str", required=True),
                        format=dict(type="str", choices=["CSV", "ORC", "Parquet"], required=True),
                        prefix=dict(type="str"),
                    ),
                    required=True,
                ),
                filter=dict(type="str"),
                optional_fields=dict(
                    type="list",
                    elements="str",
                    choices=[
                        "Size",
                        "LastModifiedDate",
                        "StorageClass",
                        "ETag",
                        "IsMultipartUploaded",
                        "ReplicationStatus",
                        "EncryptionStatus",
                        "ObjectLockRetainUntilDate",
                        "ObjectLockMode",
                        "ObjectLockLegalHoldStatus",
                        "IntelligentTieringAccessTier",
                        "BucketKeyStatus",
                        "ChecksumAlgorithm",
                        "ObjectAccessControlList",
                        "ObjectOwner",
                    ],
                ),
                id=dict(type="str", required=True),
                schedule=dict(type="str", choices=["Daily", "Weekly"], required=True),
                included_object_versions=dict(type="str", choices=["All", "Current"], required=True),
            ),
        ),
        logging=dict(
            type="dict",
            options=dict(
                target_bucket=dict(type="str", required=True),
                target_prefix=dict(type="str", default=""),
            ),
        ),
    )

    required_by = dict(
        encryption_key_id=("encryption",),
        object_lock_default_retention=("object_lock_enabled",),
    )

    mutually_exclusive = [
        ["public_access", "delete_public_access"],
        ["delete_object_ownership", "object_ownership"],
        ["dualstack", "endpoint_url"],
    ]

    required_if = [
        ["ceph", True, ["endpoint_url"]],
    ]

    module = AnsibleAWSModule(
        argument_spec=argument_spec,
        required_by=required_by,
        required_if=required_if,
        mutually_exclusive=mutually_exclusive,
    )

    # Parameter validation
    encryption = module.params.get("encryption")
    encryption_key_id = module.params.get("encryption_key_id")
    if encryption_key_id is not None and encryption != "aws:kms":
        module.fail_json(
            msg="Only 'aws:kms' is a valid option for encryption parameter when you specify encryption_key_id."
        )

    extra_params = s3_extra_params(module.params)
    retry_decorator = AWSRetry.jittered_backoff(
        max_delay=120,
        catch_extra_error_codes=["NoSuchBucket", "OperationAborted"],
    )
    s3_client = module.client("s3", retry_decorator=retry_decorator, **extra_params)

    if module.params.get("validate_bucket_name"):
        err = validate_bucket_name(module.params["name"])
        if err:
            module.fail_json(msg=err)

    account_regional = module.params.get("account_regional")
    if account_regional:
        region = module.params.get("region") or module.region
        err = validate_account_regional_bucket_name(module.params["name"], region)
        if err:
            module.fail_json(msg=err)

    state = module.params.get("state")

    try:
        if state == "present":
            create_or_update_bucket(s3_client, module)
        elif state == "absent":
            destroy_bucket(s3_client, module)
    except AnsibleS3Error as e:
        module.fail_json_aws_error(e)


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2023, Shreyas Srish <ssrish@cisco.com>
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

ANSIBLE_METADATA = {"metadata_version": "1.1", "status": ["preview"], "supported_by": "community"}

DOCUMENTATION = r"""
---
module: aci_l3out_floating_svi_path
short_description: Manage Layer 3 Outside (L3Out) Floating SVI Path Attributes (l3ext:RsDynPathAtt)
description:
- Manages L3Out Floating SVI path attributes on Cisco ACI fabrics.
options:
  tenant:
    description:
    - Name of an existing tenant.
    type: str
    aliases: [ tenant_name ]
    required: true
  l3out:
    description:
    - Name of an existing L3Out.
    type: str
    aliases: [ l3out_name ]
    required: true
  node_profile:
    description:
    - Name of the node profile.
    type: str
    aliases: [ node_profile_name, logical_node ]
    required: true
  interface_profile:
    description:
    - Name of the interface profile.
    type: str
    aliases: [ interface_profile_name, logical_interface ]
    required: true
  pod_id:
    description:
    - Pod to build the interface on.
    type: str
    required: true
  node_id:
    description:
    - Node to build the interface on for Port-channels and single ports.
    type: str
    required: true
  encap:
    description:
    - Encapsulation on the interface (e.g. "vlan-500")
    type: str
    required: true
  domain:
    description:
    - This option allows virtual machines to send frames with a mac address.
    type: str
  domain_type:
    description:
    - The domain type of the path.
    - The physical domain type is only supported in APIC v5.0 and above.
    type: str
    choices: [ physical, vmware ]
  access_encap:
    description:
    - The port encapsulation option.
    type: str
  floating_ip:
    description:
    - The floating IP address.
    type: str
    aliases: [ floating_address ]
  forged_transmit:
    description:
    - This option allows virtual machines to send frames with a mac address.
    - This is only supported in APIC v5.0 and above.
    type: str
    choices: [ enabled, disabled ]
  mac_change:
    description:
    - The status of the mac address change support for port groups in an external VMM controller.
    - This is only supported in APIC v5.0 and above.
    type: str
    choices: [ enabled, disabled ]
  promiscuous_mode:
    description:
    - The status of promiscuous mode for port groups in an external VMM controller.
    - This is only supported in APIC v5.0 and above.
    type: str
    choices: [ enabled, disabled ]
  enhanced_lag_policy:
    description:
    - The enhanced lag policy of the path.
    - Pass "" as the value to remove an existing enhanced lag policy (See Examples).
    type: str
  state:
    description:
    - Use C(present) or C(absent) for adding or removing.
    - Use C(query) for listing an object or multiple objects.
    type: str
    choices: [ absent, present, query ]
    default: present
extends_documentation_fragment:
- cisco.aci.aci
- cisco.aci.annotation

notes:
- The C(tenant), C(l3out), C(logical_node_profile), C(logical_interface_profile) and C(floating_svi) must exist before using this module in your playbook.
  The M(cisco.aci.aci_tenant), M(cisco.aci.aci_l3out), M(cisco.aci.aci_l3out_logical_node_profile), M(cisco.aci.aci_l3out_logical_interface_profile) and
  M(cisco.aci.aci_l3out_floating_svi) can be used for this.
seealso:
- module: cisco.aci.aci_tenant
- module: cisco.aci.aci_l3out
- module: cisco.aci.aci_l3out_logical_node_profile
- module: cisco.aci.aci_l3out_logical_interface_profile
- module: cisco.aci.aci_l3out_floating_svi
- name: APIC Management Information Model reference
  description: More information about the internal APIC class B(l3ext:RsDynPathAtt))
  link: https://developer.cisco.com/docs/apic-mim-ref/
author:
- Shreyas Srish (@shrsr)
"""

EXAMPLES = r"""
- name: Create a Floating SVI path attribute
  cisco.aci.aci_l3out_floating_svi_path:
    host: apic
    username: admin
    password: SomeSecretPassword
    tenant: my_tenant
    l3out: my_l3out
    node_profile: my_node_profile
    interface_profile: my_interface_profile
    pod_id: 1
    node_id: 201
    encap: vlan-1
    floating_ip: 23.45.67.90/24
    domain_type: virtual
    domain: anstest
    enhanced_lag_policy: enhanced
    state: present
  delegate_to: localhost

- name: Remove enhanced lag policy from the path
  cisco.aci.aci_l3out_floating_svi_path:
    host: apic
    username: admin
    password: SomeSecretPassword
    tenant: my_tenant
    l3out: my_l3out
    node_profile: my_node_profile
    interface_profile: my_interface_profile
    pod_id: 1
    node_id: 201
    encap: vlan-1
    floating_ip: 23.45.67.90/24
    domain_type: virtual
    domain: anstest
    enhanced_lag_policy: ""
    state: present
  delegate_to: localhost

- name: Remove a Floating SVI path attribute
  cisco.aci.aci_l3out_floating_svi_path:
    host: apic
    username: admin
    password: SomeSecretPassword
    tenant: my_tenant
    l3out: my_l3out
    node_profile: my_node_profile
    interface_profile: my_interface_profile
    pod_id: 1
    node_id: 201
    encap: vlan-1
    domain_type: virtual
    domain: anstest
    state: absent
  delegate_to: localhost

- name: Query a Floating SVI path attribute
  cisco.aci.aci_l3out_floating_svi_path:
    host: apic
    username: admin
    password: SomeSecretPassword
    tenant: my_tenant
    l3out: my_l3out
    node_profile: my_node_profile
    interface_profile: my_interface_profile
    pod_id: 1
    node_id: 201
    encap: vlan-1
    domain_type: virtual
    domain: anstest
    state: query
  delegate_to: localhost
  register: query_result

- name: Query all the Floating SVI path attributes
  cisco.aci.aci_l3out_floating_svi_path:
    host: apic
    username: admin
    password: SomeSecretPassword
    tenant: my_tenant
    l3out: my_l3out
    node_profile: my_node_profile
    interface_profile: my_interface_profile
    pod_id: 1
    node_id: 201
    encap: vlan-1
    state: query
  delegate_to: localhost
  register: query_results
"""

RETURN = r"""
current:
  description: The existing configuration from the APIC after the module has finished
  returned: success
  type: list
  sample:
    [
        {
            "fvTenant": {
                "attributes": {
                    "descr": "Production environment",
                    "dn": "uni/tn-production",
                    "name": "production",
                    "nameAlias": "",
                    "ownerKey": "",
                    "ownerTag": ""
                }
            }
        }
    ]
error:
  description: The error information as returned from the APIC
  returned: failure
  type: dict
  sample:
    {
        "code": "122",
        "text": "unknown managed object class foo"
    }
raw:
  description: The raw output returned by the APIC REST API (xml or json)
  returned: parse error
  type: str
  sample: '<?xml version="1.0" encoding="UTF-8"?><imdata totalCount="1"><error code="122" text="unknown managed object class foo"/></imdata>'
sent:
  description: The actual/minimal configuration pushed to the APIC
  returned: info
  type: list
  sample:
    {
        "fvTenant": {
            "attributes": {
                "descr": "Production environment"
            }
        }
    }
previous:
  description: The original configuration from the APIC before the module has started
  returned: info
  type: list
  sample:
    [
        {
            "fvTenant": {
                "attributes": {
                    "descr": "Production",
                    "dn": "uni/tn-production",
                    "name": "production",
                    "nameAlias": "",
                    "ownerKey": "",
                    "ownerTag": ""
                }
            }
        }
    ]
proposed:
  description: The assembled configuration from the user-provided parameters
  returned: info
  type: dict
  sample:
    {
        "fvTenant": {
            "attributes": {
                "descr": "Production environment",
                "name": "production"
            }
        }
    }
filter_string:
  description: The filter string used for the request
  returned: failure or debug
  type: str
  sample: ?rsp-prop-include=config-only
method:
  description: The HTTP method used for the request to the APIC
  returned: failure or debug
  type: str
  sample: POST
response:
  description: The HTTP response from the APIC
  returned: failure or debug
  type: str
  sample: OK (30 bytes)
status:
  description: The HTTP status from the APIC
  returned: failure or debug
  type: int
  sample: 200
url:
  description: The HTTP url used for the request to the APIC
  returned: failure or debug
  type: str
  sample: https://10.11.12.13/api/mo/uni/tn-production.json
"""


from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.aci.plugins.module_utils.aci import ACIModule, aci_argument_spec, aci_annotation_spec


def main():
    argument_spec = aci_argument_spec()
    argument_spec.update(aci_annotation_spec())
    argument_spec.update(
        tenant=dict(type="str", aliases=["tenant_name"], required=True),
        l3out=dict(type="str", aliases=["l3out_name"], required=True),
        node_profile=dict(type="str", aliases=["node_profile_name", "logical_node"], required=True),
        interface_profile=dict(type="str", aliases=["interface_profile_name", "logical_interface"], required=True),
        state=dict(type="str", default="present", choices=["absent", "present", "query"]),
        pod_id=dict(type="str", required=True),
        node_id=dict(type="str", required=True),
        encap=dict(type="str", required=True),
        floating_ip=dict(type="str", aliases=["floating_address"]),
        forged_transmit=dict(type="str", choices=["enabled", "disabled"]),
        mac_change=dict(type="str", choices=["enabled", "disabled"]),
        promiscuous_mode=dict(type="str", choices=["enabled", "disabled"]),
        domain_type=dict(type="str", choices=["physical", "vmware"]),
        domain=dict(type="str"),
        enhanced_lag_policy=dict(type="str"),
        access_encap=dict(type="str"),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_if=[
            ["state", "present", ["domain_type", "domain", "floating_ip"]],
            ["state", "absent", ["domain_type", "domain"]],
        ],
    )

    tenant = module.params.get("tenant")
    l3out = module.params.get("l3out")
    node_profile = module.params.get("node_profile")
    interface_profile = module.params.get("interface_profile")
    state = module.params.get("state")
    pod_id = module.params.get("pod_id")
    node_id = module.params.get("node_id")
    floating_ip = module.params.get("floating_ip")
    encap = module.params.get("encap")
    forged_transmit = module.params.get("forged_transmit").capitalize() if module.params.get("forged_transmit") else None
    mac_change = module.params.get("mac_change").capitalize() if module.params.get("mac_change") else None
    promiscuous_mode = module.params.get("promiscuous_mode").capitalize() if module.params.get("promiscuous_mode") else None
    domain_type = module.params.get("domain_type")
    domain = module.params.get("domain")
    enhanced_lag_policy = module.params.get("enhanced_lag_policy")
    access_encap = module.params.get("access_encap")

    aci = ACIModule(module)

    node_dn = "topology/pod-{0}/node-{1}".format(pod_id, node_id)

    tDn = None
    if domain_type == "physical":
        tDn = "uni/phys-{0}".format(domain)
    elif domain_type == "vmware":
        tDn = "uni/vmmp-VMware/dom-{0}".format(domain)

    aci.construct_url(
        root_class=dict(
            aci_class="fvTenant",
            aci_rn="tn-{0}".format(tenant),
            module_object=tenant,
            target_filter={"name": tenant},
        ),
        subclass_1=dict(
            aci_class="l3extOut",
            aci_rn="out-{0}".format(l3out),
            module_object=l3out,
            target_filter={"name": l3out},
        ),
        subclass_2=dict(
            aci_class="l3extLNodeP",
            aci_rn="lnodep-{0}".format(node_profile),
            module_object=node_profile,
            target_filter={"name": node_profile},
        ),
        subclass_3=dict(
            aci_class="l3extLIfP",
            aci_rn="lifp-{0}".format(interface_profile),
            module_object=interface_profile,
            target_filter={"name": interface_profile},
        ),
        subclass_4=dict(
            aci_class="l3extVirtualLIfP", aci_rn="vlifp-[{0}]-[{1}]".format(node_dn, encap), module_object=node_dn, target_filter={"nodeDn": node_dn}
        ),
        subclass_5=dict(
            aci_class="l3extRsDynPathAtt",
            aci_rn="rsdynPathAtt-[{0}]".format(tDn),
            module_object=tDn,
            target_filter={"tDn": tDn},
        ),
        child_classes=["l3extVirtualLIfPLagPolAtt"],
    )

    aci.get_existing()

    if state == "present":
        child_configs = []
        if enhanced_lag_policy is not None and domain_type == "vmware":
            existing_enhanced_lag_policy = ""
            if isinstance(aci.existing, list) and len(aci.existing) > 0:
                for child in aci.existing[0].get("l3extRsDynPathAtt", {}).get("children", {}):
                    if child.get("l3extVirtualLIfPLagPolAtt"):
                        try:
                            existing_enhanced_lag_policy = child["l3extVirtualLIfPLagPolAtt"]["children"][0]["l3extRsVSwitchEnhancedLagPol"]["attributes"][
                                "tDn"
                            ].split("enlacplagp-")[1]
                        except (AttributeError, IndexError, KeyError):
                            existing_enhanced_lag_policy = ""

                        if enhanced_lag_policy == "":
                            child_configs.append(
                                dict(
                                    l3extVirtualLIfPLagPolAtt=dict(
                                        attributes=dict(status="deleted"),
                                    ),
                                )
                            )

            if enhanced_lag_policy != "":
                child = [
                    dict(
                        l3extRsVSwitchEnhancedLagPol=dict(
                            attributes=dict(tDn="{0}/vswitchpolcont/enlacplagp-{1}".format(tDn, enhanced_lag_policy)),
                        )
                    ),
                ]
                if enhanced_lag_policy != existing_enhanced_lag_policy and existing_enhanced_lag_policy != "":
                    child.append(
                        dict(
                            l3extRsVSwitchEnhancedLagPol=dict(
                                attributes=dict(status="deleted", tDn="{0}/vswitchpolcont/enlacplagp-{1}".format(tDn, existing_enhanced_lag_policy)),
                            )
                        )
                    )
                child_configs.append(dict(l3extVirtualLIfPLagPolAtt=dict(attributes=dict(), children=child)))

        aci.payload(
            aci_class="l3extRsDynPathAtt",
            class_config=dict(
                floatingAddr=floating_ip,
                forgedTransmit=forged_transmit,
                macChange=mac_change,
                promMode=promiscuous_mode,
                encap=access_encap,
            ),
            child_configs=child_configs,
        )

        aci.get_diff(aci_class="l3extRsDynPathAtt")

        aci.post_config()

    elif state == "absent":
        aci.delete_config()

    aci.exit_json()


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# Copyright 2021 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function


__metaclass__ = type

"""
The Snmp_server parser templates file. This contains
a list of parser definitions and associated functions that
facilitates both facts gathering and native command generation for
the given network resource.
"""

import re

from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.rm_base.network_template import (
    NetworkTemplate,
)


def _template_hosts(data):
    cmd = "snmp-server host {0}".format(data["host"])
    if data.get("traps"):
        cmd += " traps"
    if data.get("informs"):
        cmd += " informs"
    if data.get("use_vrf"):
        cmd += " use-vrf {0}".format(data["use_vrf"])
    if data.get("filter_vrf"):
        cmd += " filter-vrf {0}".format(data["filter_vrf"])
    if data.get("source_interface"):
        cmd += " source-interface {0}".format(data["source_interface"])
    if data.get("version"):
        cmd += " version {0}".format(data["version"])
    if data.get("community"):
        cmd += " " + data["community"]
    elif data.get("auth"):
        cmd += " auth {0}".format(data["auth"])
    elif data.get("priv"):
        cmd += " priv {0}".format(data["priv"])
    if data.get("udp_port"):
        cmd += " udp-port {0}".format(data["udp_port"])

    return cmd


def _tmplt_users_auth(data):
    cmd = "snmp-server user {0}".format(data["user"])

    if "group" in data:
        cmd += " {0}".format(data["group"])
    if "authentication" in data:
        auth = data["authentication"]
        if "algorithm" in auth:
            cmd += " auth {0}".format(auth["algorithm"])
        if "password" in auth:
            cmd += " {0}".format(auth["password"])
        priv = auth.get("priv", {})
        if priv:
            cmd += " priv"
            if priv.get("aes_128", False):
                cmd += " aes-128"
            if "privacy_password" in priv:
                cmd += " {0}".format(priv["privacy_password"])
        if auth.get("localized_key", False):
            cmd += " localizedkey"
        elif auth.get("localizedv2_key", False):
            cmd += " localizedV2key"
        if "engine_id" in auth:
            cmd += " engineID {0}".format(auth["engine_id"])

        return cmd


class Snmp_serverTemplate(NetworkTemplate):
    def __init__(self, lines=None, module=None):
        super(Snmp_serverTemplate, self).__init__(
            lines=lines,
            tmplt=self,
            module=module,
        )

    # fmt: off
    PARSERS = [
        {
            "name": "aaa_user.cache_timeout",
            "getval": re.compile(
                r"""
                ^snmp-server\saaa-user
                \scache-timeout\s(?P<cache_timeout>\d+)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server aaa-user cache-timeout {{ aaa_user.cache_timeout }}",
            "result": {
                "aaa_user": {
                    "cache_timeout": "{{ cache_timeout }}",
                },
            },
        },
        {
            "name": "communities",
            "getval": re.compile(
                r"""
                ^snmp-server
                \scommunity\s(?P<community>\S+)
                (\sgroup\s(?P<group>\S+))?
                (\suse-ipv4acl\s(?P<use_ipv4acl>\S+))?
                (\suse-ipv6acl\s(?P<use_ipv6acl>\S+))?
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server community {{ name }} {{ (' group ' + group) if group is defined else '' }} \n"
                      "snmp-server community {{ name }} {{ ' ro' if ro|d(False) else ''}}"
                      "{{ ' rw' if rw|d(False) else ''}} \n"
                      "snmp-server community {{ name }}"
                      "{{ (' use-ipv4acl ' + use_ipv4acl) if use_ipv4acl is defined else '' }} "
                      "{{ (' use-ipv6acl ' + use_ipv6acl) if use_ipv6acl is defined else '' }}",

            "result": {
                "communities": [
                    {
                        "name": "{{ community }}",
                        "group": "{{ group }}",
                        "use_ipv4acl": "{{ use_ipv4acl }}",
                        "use_ipv6acl": "{{ use_ipv6acl }}",
                    },
                ],
            },
        },
        {
            "name": "contact",
            "getval": re.compile(
                r"""
                ^snmp-server
                \scontact\s(?P<contact>.+)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server contact {{ contact }}",
            "result": {
                "contact": "{{ contact }}",
            },
        },
        {
            "name": "context",
            "getval": re.compile(
                r"""
                ^snmp-server
                \scontext\s(?P<name>\S+)
                (\sinstance\s(?P<instance>\S+))?
                (\svrf\s(?P<vrf>\S+))?
                (\stopology\s(?P<topology>\S+))?
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server context {{ context.name }}"
                      "{{ ' instance ' + context.instance if context.instance is defined else '' }}"
                      "{{ ' topology ' + context.topology if context.topology is defined else '' }}"
                      "{{ ' vrf ' + context.vrf if context.vrf is defined else '' }}",
            "result": {
                "context": {
                    "name": "{{ name }}",
                    "instance": "{{ instance }}",
                    "vrf": "{{ vrf }}",
                    "topology": "{{ topology }}",
                },

            },
        },
        {
            "name": "counter.enable",
            "getval": re.compile(
                r"""
                ^snmp-server
                \scounter
                \scache\s(?P<enable>enable)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server counter cache enable",
            "result": {
                "counter": {
                    "cache": {
                        "enable": "{{ True if enable is defined else None }}",
                    },
                },
            },
        },
        {
            "name": "counter.cache.timeout",
            "getval": re.compile(
                r"""
                ^snmp-server
                \scounter
                \scache\stimeout\s(?P<timeout>\d+)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server counter cache timeout {{ counter.cache.timeout }}",
            "result": {
                "counter": {
                    "cache": {
                        "timeout": "{{ timeout }}",
                    },
                },
            },
        },
        {
            "name": "drop.unknown_engine_id",
            "getval": re.compile(
                r"""
                ^snmp-server\sdrop
                \s(?P<unknown_engine_id>unknown-engine-id)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server drop unknown-engine-id",
            "result": {
                "drop": {
                    "unknown_engine_id": "{{ not not unknown_engine_id }}",
                },
            },
        },
        {
            "name": "drop.unknown_user",
            "getval": re.compile(
                r"""
                ^snmp-server\sdrop
                \s(?P<unknown_user>unknown-user)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server drop unknown-user",
            "result": {
                "drop": {
                    "unknown_user": "{{ not not unknown_user }}",
                },
            },
        },
        {
            "name": "traps.aaa",
            "getval": re.compile(
                r"""
                ^snmp-server\senable
                \straps\saaa\s(?P<server_state_change>server-state-change)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server enable traps aaa"
                      "{{ ' server-state-change' if traps.aaa.server_state_change|d(False) else ''}}",
            "result": {
                "traps": {
                    "aaa": {
                        "server_state_change": "{{ not not server_state_change }}",
                    },
                },
            },
        },
        {
            "name": "traps.bgp",
            "getval": re.compile(
                r"""
                ^snmp-server\senable
                \straps\s(?P<enable>bgp)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server enable traps bgp",
            "result": {
                "traps": {
                    "bgp": {
                        "enable": "{{ not not enable }}",
                    },
                },
            },
        },
        {
            "name": "traps.bridge.newroot",
            "getval": re.compile(
                r"""
                ^snmp-server\senable
                \straps
                \sbridge\s(?P<newroot>newroot)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server enable traps bridge newroot",
            "result": {
                "traps": {
                    "bridge": {
                        "newroot": "{{ not not newroot }}",
                    },
                },
            },
        },
        {
            "name": "traps.bridge.topologychange",
            "getval": re.compile(
                r"""
                ^snmp-server\senable
                \straps
                \sbridge\s(?P<topologychange>topologychange)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server enable traps bridge topologychange",
            "result": {
                "traps": {
                    "bridge": {
                        "topologychange": "{{ not not topologychange }}",
                    },
                },
            },
        },
        {
            "name": "traps.callhome.event_notify",
            "getval": re.compile(
                r"""
                ^snmp-server\senable
                \straps
                \scallhome\s(?P<event_notify>event-notify)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server enable traps callhome event-notify",
            "result": {
                "traps": {
                    "callhome": {
                        "event_notify": "{{ not not event_notify }}",
                    },
                },
            },
        },
        {
            "name": "traps.callhome.smtp_send_fail",
            "getval": re.compile(
                r"""
                ^snmp-server\senable
                \straps
                \scallhome\s(?P<smtp_send_fail>smtp-send-fail)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server enable traps callhome smtp-send-fail",
            "result": {
                "traps": {
                    "callhome": {
                        "smtp_send_fail": "{{ not not smtp_send_fail }}",
                    },
                },
            },
        },
        {
            "name": "traps.cfs.merge_failure",
            "getval": re.compile(
                r"""
                ^snmp-server\senable
                \straps
                \scfs\s(?P<merge_failure>merge-failure)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server enable traps cfs merge-failure",
            "result": {
                "traps": {
                    "cfs": {
                        "merge_failure": "{{ not not merge_failure }}",
                    },
                },
            },
        },
        {
            "name": "traps.cfs.state_change_notif",
            "getval": re.compile(
                r"""
                ^snmp-server\senable
                \straps
                \scfs\s(?P<state_change_notif>state-change-notif)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server enable traps cfs state-change-notif",
            "result": {
                "traps": {
                    "cfs": {
                        "state_change_notif": "{{ not not state_change_notif }}",
                    },
                },
            },
        },
        {
            "name": "traps.config.ccmCLIRunningConfigChanged",
            "getval": re.compile(
                r"""
                ^snmp-server\senable
                \straps
                \sconfig\s(?P<ccmCLIRunningConfigChanged>ccmCLIRunningConfigChanged)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server enable traps config ccmCLIRunningConfigChanged",
            "result": {
                "traps": {
                    "config": {
                        "ccmCLIRunningConfigChanged": "{{ not not ccmCLIRunningConfigChanged }}",
                    },
                },
            },
        },
        {
            "name": "traps.entity.cefcMIBEnableStatusNotification",
            "getval": re.compile(
                r"""
                ^snmp-server\senable
                \straps
                \sentity\s(?P<cefcMIBEnableStatusNotification>cefcMIBEnableStatusNotification)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server enable traps entity cefcMIBEnableStatusNotification",
            "result": {
                "traps": {
                    "entity": {
                        "cefcMIBEnableStatusNotification": "{{ not not cefcMIBEnableStatusNotification }}",
                    },
                },
            },
        },
        {
            "name": "traps.entity.entity_fan_status_change",
            "getval": re.compile(
                r"""
                ^snmp-server\senable
                \straps
                \sentity\s(?P<entity_fan_status_change>entity_fan_status_change)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server enable traps entity entity_fan_status_change",
            "result": {
                "traps": {
                    "entity": {
                        "entity_fan_status_change": "{{ not not entity_fan_status_change }}",
                    },
                },
            },
        },
        {
            "name": "traps.entity.entity_mib_change",
            "getval": re.compile(
                r"""
                ^snmp-server\senable
                \straps
                \sentity\s(?P<entity_mib_change>entity_mib_change)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server enable traps entity entity_mib_change",
            "result": {
                "traps": {
                    "entity": {
                        "entity_mib_change": "{{ not not entity_mib_change }}",
                    },
                },
            },
        },
        {
            "name": "traps.entity.entity_module_inserted",
            "getval": re.compile(
                r"""
                ^snmp-server\senable
                \straps
                \sentity\s(?P<entity_module_inserted>entity_module_inserted)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server enable traps entity entity_module_inserted",
            "result": {
                "traps": {
                    "entity": {
                        "entity_module_inserted": "{{ not not entity_module_inserted }}",
                    },
                },
            },
        },
        {
            "name": "traps.entity.entity_module_status_change",
            "getval": re.compile(
                r"""
                ^snmp-server\senable
                \straps
                \sentity\s(?P<entity_module_status_change>entity_module_status_change)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server enable traps entity entity_module_status_change",
            "result": {
                "traps": {
                    "entity": {
                        "entity_module_status_change": "{{ not not entity_module_status_change }}",
                    },
                },
            },
        },
        {
            "name": "traps.entity.entity_power_out_change",
            "getval": re.compile(
                r"""
                ^snmp-server\senable
                \straps
                \sentity\s(?P<entity_power_out_change>entity_power_out_change)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server enable traps entity entity_power_out_change",
            "result": {
                "traps": {
                    "entity": {
                        "entity_power_out_change": "{{ not not entity_power_out_change }}",
                    },
                },
            },
        },
        {
            "name": "traps.entity.entity_power_status_change",
            "getval": re.compile(
                r"""
                ^snmp-server\senable
                \straps
                \sentity\s(?P<entity_power_status_change>entity_power_status_change)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server enable traps entity entity_power_status_change",
            "result": {
                "traps": {
                    "entity": {
                        "entity_power_status_change": "{{ not not entity_power_status_change }}",
                    },
                },
            },
        },
        {
            "name": "traps.entity.entity_sensor",
            "getval": re.compile(
                r"""
                ^snmp-server\senable
                \straps
                \sentity\s(?P<entity_sensor>entity_sensor)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server enable traps entity entity_sensor",
            "result": {
                "traps": {
                    "entity": {
                        "entity_sensor": "{{ not not entity_sensor }}",
                    },
                },
            },
        },
        {
            "name": "traps.entity.entity_unrecognised_module",
            "getval": re.compile(
                r"""
                ^snmp-server\senable
                \straps
                \sentity\s(?P<entity_unrecognised_module>entity_unrecognised_module)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server enable traps entity entity_unrecognised_module",
            "result": {
                "traps": {
                    "entity": {
                        "entity_unrecognised_module": "{{ not not entity_unrecognised_module }}",
                    },
                },
            },
        },
        {
            "name": "traps.feature_control.featureOpStatusChange",
            "getval": re.compile(
                r"""
                ^snmp-server\senable
                \straps
                \sfeature-control\s(?P<featureOpStatusChange>featureOpStatusChange)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server enable traps feature-control featureOpStatusChange",
            "result": {
                "traps": {
                    "feature_control": {
                        "featureOpStatusChange": "{{ not not featureOpStatusChange }}",
                    },
                },
            },
        },
        {
            "name": "traps.feature_control.ciscoFeatOpStatusChange",
            "getval": re.compile(
                r"""
                ^snmp-server\senable
                \straps
                \sfeature-control\s(?P<ciscoFeatOpStatusChange>ciscoFeatOpStatusChange)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server enable traps feature-control ciscoFeatOpStatusChange",
            "result": {
                "traps": {
                    "feature_control": {
                        "ciscoFeatOpStatusChange": "{{ not not ciscoFeatOpStatusChange }}",
                    },
                },
            },
        },
        {
            "name": "traps.generic.coldStart",
            "getval": re.compile(
                r"""
                ^snmp-server\senable
                \straps
                \sgeneric\s(?P<coldStart>coldStart)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server enable traps generic coldStart",
            "result": {
                "traps": {
                    "generic": {
                        "coldStart": "{{ not not coldStart }}",
                    },
                },
            },
        },
        {
            "name": "traps.generic.warmStart",
            "getval": re.compile(
                r"""
                ^snmp-server\senable
                \straps
                \sgeneric\s(?P<warmStart>warmStart)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server enable traps generic warmStart",
            "result": {
                "traps": {
                    "generic": {
                        "warmStart": "{{ not not warmStart }}",
                    },
                },
            },
        },
        {
            "name": "traps.license.notify_license_expiry",
            "getval": re.compile(
                r"""
                ^snmp-server\senable
                \straps
                \slicense\s(?P<notify_license_expiry>notify_license_expiry)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server enable traps license notify-license-expiry",
            "result": {
                "traps": {
                    "license": {
                        "notify_license_expiry": "{{ not not notify_license_expiry }}",
                    },
                },
            },
        },
        {
            "name": "traps.license.notify_license_expiry_warning",
            "getval": re.compile(
                r"""
                ^snmp-server\senable
                \straps
                \slicense\s(?P<notify_license_expiry_warning>notify-license-expiry-warning)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server enable traps license notify-license-expiry-warning",
            "result": {
                "traps": {
                    "license": {
                        "notify_license_expiry_warning": "{{ not not notify_license_expiry_warning }}",
                    },
                },
            },
        },
        {
            "name": "traps.license.notify_licensefile_missing",
            "getval": re.compile(
                r"""
                ^snmp-server\senable
                \straps
                \slicense\s(?P<notify_licensefile_missing>notify-licensefile-missing)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server enable traps license notify-licensefile-missing",
            "result": {
                "traps": {
                    "license": {
                        "notify_licensefile_missing": "{{ not not notify_licensefile_missing }}",
                    },
                },
            },
        },
        {
            "name": "traps.license.notify_no_license_for_feature",
            "getval": re.compile(
                r"""
                ^snmp-server\senable
                \straps
                \slicense\s(?P<notify_no_license_for_feature>notify-no-license-for-feature)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server enable traps license notify-no-license-for-feature",
            "result": {
                "traps": {
                    "license": {
                        "notify_no_license_for_feature": "{{ not not notify_no_license_for_feature }}",
                    },
                },
            },
        },
        {
            "name": "traps.link.cErrDisableInterfaceEventRev1",
            "getval": re.compile(
                r"""
                ^snmp-server\senable
                \straps
                \slink\s(?P<cErrDisableInterfaceEventRev1>cErrDisableInterfaceEventRev1)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server enable traps link cErrDisableInterfaceEventRev1",
            "result": {
                "traps": {
                    "link": {
                        "cErrDisableInterfaceEventRev1": "{{ not not cErrDisableInterfaceEventRev1 }}",
                    },
                },
            },
        },
        {
            "name": "traps.link.cieLinkDown",
            "getval": re.compile(
                r"""
                ^snmp-server\senable
                \straps
                \slink\s(?P<cieLinkDown>cieLinkDown)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server enable traps link cieLinkDown",
            "result": {
                "traps": {
                    "link": {
                        "cieLinkDown": "{{ not not cieLinkDown }}",
                    },
                },
            },
        },
        {
            "name": "traps.link.cieLinkUp",
            "getval": re.compile(
                r"""
                ^snmp-server\senable
                \straps
                \slink\s(?P<cieLinkUp>cieLinkUp)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server enable traps link cieLinkUp",
            "result": {
                "traps": {
                    "link": {
                        "cieLinkUp": "{{ not not cieLinkUp }}",
                    },
                },
            },
        },
        {
            "name": "traps.link.cisco_xcvr_mon_status_chg",
            "getval": re.compile(
                r"""
                ^snmp-server\senable
                \straps
                \slink\s(?P<cisco_xcvr_mon_status_chg>cisco-xcvr-mon-status-chg)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server enable traps link cisco-xcvr-mon-status-chg",
            "result": {
                "traps": {
                    "link": {
                        "cisco_xcvr_mon_status_chg": "{{ not not cisco_xcvr_mon_status_chg }}",
                    },
                },
            },
        },
        {
            "name": "traps.link.cmn_mac_move_notification",
            "getval": re.compile(
                r"""
                ^snmp-server\senable
                \straps
                \slink\s(?P<cmn_mac_move_notification>cmn-mac-move-notification)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server enable traps link cmn-mac-move-notification",
            "result": {
                "traps": {
                    "link": {
                        "cmn_mac_move_notification": "{{ not not cmn_mac_move_notification }}",
                    },
                },
            },
        },
        {
            "name": "traps.link.delayed_link_state_change",
            "getval": re.compile(
                r"""
                ^snmp-server\senable
                \straps
                \slink\s(?P<delayed_link_state_change>delayed-link-state-change)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server enable traps link delayed-link-state-change",
            "result": {
                "traps": {
                    "link": {
                        "delayed_link_state_change": "{{ not not delayed_link_state_change }}",
                    },
                },
            },
        },
        {
            "name": "traps.link.extended_linkDown",
            "getval": re.compile(
                r"""
                ^snmp-server\senable
                \straps
                \slink\s(?P<extended_linkDown>extended-linkDown)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server enable traps link extended-linkDown",
            "result": {
                "traps": {
                    "link": {
                        "extended_linkDown": "{{ not not extended_linkDown }}",
                    },
                },
            },
        },
        {
            "name": "traps.link.extended_linkUp",
            "getval": re.compile(
                r"""
                ^snmp-server\senable
                \straps
                \slink\s(?P<extended_linkUp>extended-linkUp)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server enable traps link extended-linkUp",
            "result": {
                "traps": {
                    "link": {
                        "extended_linkUp": "{{ not not extended_linkUp }}",
                    },
                },
            },
        },
        {
            "name": "traps.link.linkDown",
            "getval": re.compile(
                r"""
                ^snmp-server\senable
                \straps
                \slink\s(?P<linkDown>linkDown)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server enable traps link linkDown",
            "result": {
                "traps": {
                    "link": {
                        "linkDown": "{{ not not linkDown }}",
                    },
                },
            },
        },
        {
            "name": "traps.link.linkUp",
            "getval": re.compile(
                r"""
                ^snmp-server\senable
                \straps
                \slink\s(?P<linkUp>linkUp)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server enable traps link linkUp",
            "result": {
                "traps": {
                    "link": {
                        "linkUp": "{{ not not linkUp }}",
                    },
                },
            },
        },
        {
            "name": "traps.mmode.cseMaintModeChangeNotify",
            "getval": re.compile(
                r"""
                ^snmp-server\senable
                \straps
                \smmode\s(?P<cseMaintModeChangeNotify>cseMaintModeChangeNotify)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server enable traps mmode cseMaintModeChangeNotify",
            "result": {
                "traps": {
                    "mmode": {
                        "cseMaintModeChangeNotify": "{{ not not cseMaintModeChangeNotify }}",
                    },
                },
            },
        },
        {
            "name": "traps.mmode.cseNormalModeChangeNotify",
            "getval": re.compile(
                r"""
                ^snmp-server\senable
                \straps
                \smmode\s(?P<cseNormalModeChangeNotify>cseNormalModeChangeNotify)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server enable traps mmode cseNormalModeChangeNotify",
            "result": {
                "traps": {
                    "mmode": {
                        "cseNormalModeChangeNotify": "{{ not not cseNormalModeChangeNotify }}",
                    },
                },
            },
        },
        {
            "name": "traps.ospf",
            "getval": re.compile(
                r"""
                ^snmp-server\senable
                \straps\s(?P<enable>ospf)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server enable traps ospf",
            "result": {
                "traps": {
                    "ospf": {
                        "enable": "{{ not not enable }}",
                    },
                },
            },
        },
        {
            "name": "traps.ospfv3",
            "getval": re.compile(
                r"""
                ^snmp-server\senable
                \straps\s(?P<enable>ospfv3)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server enable traps ospfv3",
            "result": {
                "traps": {
                    "ospfv3": {
                        "enable": "{{ not not enable }}",
                    },
                },
            },
        },
        {
            "name": "traps.rf.redundancy_framework",
            "getval": re.compile(
                r"""
                ^snmp-server\senable
                \straps
                \srf\s(?P<redundancy_framework>redundancy-framework)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server enable traps rf redundancy-framework",
            "result": {
                "traps": {
                    "rf": {
                        "redundancy_framework": "{{ not not redundancy_framework }}",
                    },
                },
            },
        },
        {
            "name": "traps.rmon.fallingAlarm",
            "getval": re.compile(
                r"""
                ^snmp-server\senable
                \straps
                \srmon\s(?P<fallingAlarm>fallingAlarm)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server enable traps rmon fallingAlarm",
            "result": {
                "traps": {
                    "rmon": {
                        "fallingAlarm": "{{ not not fallingAlarm }}",
                    },
                },
            },
        },
        {
            "name": "traps.rmon.hcFallingAlarm",
            "getval": re.compile(
                r"""
                ^snmp-server\senable
                \straps
                \srmon\s(?P<hcFallingAlarm>hcFallingAlarm)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server enable traps rmon hcFallingAlarm",
            "result": {
                "traps": {
                    "rmon": {
                        "hcFallingAlarm": "{{ not not hcFallingAlarm }}",
                    },
                },
            },
        },
        {
            "name": "traps.rmon.hcRisingAlarm",
            "getval": re.compile(
                r"""
                ^snmp-server\senable
                \straps
                \srmon\s(?P<hcRisingAlarm>hcRisingAlarm)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server enable traps rmon hcRisingAlarm",
            "result": {
                "traps": {
                    "rmon": {
                        "hcRisingAlarm": "{{ not not hcRisingAlarm }}",
                    },
                },
            },
        },
        {
            "name": "traps.rmon.risingAlarm",
            "getval": re.compile(
                r"""
                ^snmp-server\senable
                \straps
                \srmon\s(?P<risingAlarm>risingAlarm)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server enable traps rmon risingAlarm",
            "result": {
                "traps": {
                    "rmon": {
                        "risingAlarm": "{{ not not risingAlarm }}",
                    },
                },
            },
        },
        {
            "name": "traps.snmp.authentication",
            "getval": re.compile(
                r"""
                ^snmp-server\senable
                \straps
                \ssnmp\s(?P<authentication>authentication)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server enable traps snmp authentication",
            "result": {
                "traps": {
                    "snmp": {
                        "authentication": "{{ not not authentication }}",
                    },
                },
            },
        },
        {
            "name": "traps.storm_control.cpscEventRev1",
            "getval": re.compile(
                r"""
                ^snmp-server\senable
                \straps
                \sstorm-control\s(?P<cpscEventRev1>cpscEventRev1)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server enable traps storm-control cpscEventRev1",
            "result": {
                "traps": {
                    "storm_control": {
                        "cpscEventRev1n": "{{ not not cpscEventRev1 }}",
                    },
                },
            },
        },
        {
            "name": "traps.storm_control.trap_rate",
            "getval": re.compile(
                r"""
                ^snmp-server\senable
                \straps
                \sstorm-control\s(?P<trap_rate>trap-rate)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server enable traps storm-control trap-rate",
            "result": {
                "traps": {
                    "storm_control": {
                        "trap_rate": "{{ not not trap_rate }}",
                    },
                },
            },
        },
        {
            "name": "traps.stpx.inconsistency",
            "getval": re.compile(
                r"""
                ^snmp-server\senable
                \straps
                \sstpx\s(?P<inconsistency>inconsistency)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server enable traps stpx inconsistency",
            "result": {
                "traps": {
                    "stpx": {
                        "inconsistency": "{{ not not inconsistency }}",
                    },
                },
            },
        },
        {
            "name": "traps.stpx.root_inconsistency",
            "getval": re.compile(
                r"""
                ^snmp-server\senable
                \straps
                \sstpx\s(?P<root_inconsistency>root-inconsistency)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server enable traps stpx root-inconsistency",
            "result": {
                "traps": {
                    "stpx": {
                        "root_inconsistency": "{{ not not root_inconsistency }}",
                    },
                },
            },
        },
        {
            "name": "traps.stpx.loop_inconsistency",
            "getval": re.compile(
                r"""
                ^snmp-server\senable
                \straps
                \sstpx\s(?P<loop_inconsistency>loop-inconsistency)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server enable traps stpx loop-inconsistency",
            "result": {
                "traps": {
                    "stpx": {
                        "loop_inconsistency": "{{ not not loop_inconsistency }}",
                    },
                },
            },
        },
        {
            "name": "traps.syslog.message_generated",
            "getval": re.compile(
                r"""
                ^snmp-server\senable
                \straps
                \ssyslog\s(?P<message_generated>message-generated)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server enable traps syslog message-generated",
            "result": {
                "traps": {
                    "syslog": {
                        "message_generated": "{{ not not message_generated }}",
                    },
                },
            },
        },
        {
            "name": "traps.sysmgr.cseFailSwCoreNotifyExtended",
            "getval": re.compile(
                r"""
                ^snmp-server\senable
                \straps
                \ssysmgr\s(?P<cseFailSwCoreNotifyExtended>cseFailSwCoreNotifyExtended)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server enable traps sysmgr cseFailSwCoreNotifyExtended",
            "result": {
                "traps": {
                    "sysmgr": {
                        "cseFailSwCoreNotifyExtended": "{{ not not cseFailSwCoreNotifyExtended }}",
                    },
                },
            },
        },
        {
            "name": "traps.system.clock_change_notification",
            "getval": re.compile(
                r"""
                ^snmp-server\senable
                \straps
                \ssystem\s(?P<clock_change_notification>Clock-change-notification)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server enable traps system Clock-change-notification",
            "result": {
                "traps": {
                    "system": {
                        "clock_change_notification": "{{ not not clock_change_notification }}",
                    },
                },
            },
        },
        {
            "name": "traps.upgrade.upgradeJobStatusNotify",
            "getval": re.compile(
                r"""
                ^snmp-server\senable
                \straps
                \supgrade\s(?P<upgradeJobStatusNotify>upgradeJobStatusNotify)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server enable traps upgrade upgradeJobStatusNotify",
            "result": {
                "traps": {
                    "upgrade": {
                        "upgradeJobStatusNotify": "{{ not not upgradeJobStatusNotify }}",
                    },
                },
            },
        },
        {
            "name": "traps.upgrade.upgradeOpNotifyOnCompletion",
            "getval": re.compile(
                r"""
                ^snmp-server\senable
                \straps
                \supgrade\s(?P<upgradeOpNotifyOnCompletion>upgradeOpNotifyOnCompletion)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server enable traps upgrade upgradeOpNotifyOnCompletion",
            "result": {
                "traps": {
                    "upgrade": {
                        "upgradeOpNotifyOnCompletion": "{{ not not upgradeOpNotifyOnCompletion }}",
                    },
                },
            },
        },
        {
            "name": "traps.vtp.notifs",
            "getval": re.compile(
                r"""
                ^snmp-server\senable
                \straps
                \svtp\s(?P<notifs>notifs)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server enable traps vtp notifs",
            "result": {
                "traps": {
                    "vtp": {
                        "notifs": "{{ not not notifs }}",
                    },
                },
            },
        },
        {
            "name": "traps.vtp.vlancreate",
            "getval": re.compile(
                r"""
                ^snmp-server\senable
                \straps
                \svtp\s(?P<vlancreate>vlancreate)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server enable traps vtp vlancreate",
            "result": {
                "traps": {
                    "vtp": {
                        "vlancreate": "{{ not not vlancreate }}",
                    },
                },
            },
        },
        {
            "name": "traps.vtp.vlandelete",
            "getval": re.compile(
                r"""
                ^snmp-server\senable
                \straps
                \svtp\s(?P<vlandelete>vlandelete)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server enable traps vtp vlandelete",
            "result": {
                "traps": {
                    "vtp": {
                        "vlandelete": "{{ not not vlandelete }}",
                    },
                },
            },
        },

        {
            "name": "engine_id.local",
            "getval": re.compile(
                r"""
                ^snmp-server\sengineID
                \slocal\s(?P<local>\S+)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server engineID local {{ engine_id.local }}",
            "result": {
                "engine_id": {
                    "local": "{{ local }}",
                },
            },
        },
        {
            "name": "global_enforce_priv",
            "getval": re.compile(
                r"""
                ^snmp-server
                \s(?P<global_enforce_priv>globalEnforcePriv)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server globalEnforcePriv",
            "result": {
                "global_enforce_priv": "{{ not not global_enforce_priv }}",
            },
        },
        {
            "name": "hosts",
            "getval": re.compile(
                r"""
                ^snmp-server
                \shost\s(?P<host>\S+)
                (\s((?P<traps>traps)|(?P<informs>informs)|(use-vrf\s(?P<use_vrf>\S+)|(filter-vrf\s(?P<filter_vrf>\S+))|(source-interface\s(?P<source_interface>\S+)))))
                (\sversion\s(?P<version>\S+))?
                (\s((auth\s(?P<auth>\S+))|(priv\s(?P<priv>\S+))|((?P<community>\S+))))?
                (\sudp-port\s(?P<udp_port>\S+))?
                $""", re.VERBOSE,
            ),
            "setval": _template_hosts,
            "result": {
                "hosts": [
                    {
                        "host": "{{ host }}",
                        "community": "{{ community }}",
                        "filter_vrf": "{{ filter_vrf }}",
                        "informs": "{{ not not informs }}",
                        "source_interface": "{{ source_interface }}",
                        "traps": "{{ not not traps }}",
                        "use_vrf": "{{ use_vrf }}",
                        "version": "{{ version }}",
                        "udp_port": "{{ udp_port }}",
                        "auth": "{{ auth }}",
                        "priv": "{{ priv }}",
                    },
                ],
            },
        },
        {
            "name": "location",
            "getval": re.compile(
                r"""
                ^snmp-server
                \slocation\s(?P<location>.+)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server location {{ location }}",
            "result": {
                "location": "{{ location }}",
            },
        },
        {
            "name": "mib.community_map",
            "getval": re.compile(
                r"""
                ^snmp-server
                \smib
                \scommunity-map\s(?P<community>\S+)
                \scontext\s(?P<context>\S+)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server mib community-map {{ mib.community_map.community }} context {{ mib.community_map.context }}",
            "result": {
                "mib": {
                    "community_map": {
                        "community": "{{ community }}",
                        "context": "{{ context }}",

                    },
                },
            },
        },
        {
            "name": "packetsize",
            "getval": re.compile(
                r"""
                ^snmp-server
                \spacketsize\s(?P<packetsize>\d+)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server packetsize {{ packetsize }}",
            "result": {
                "packetsize": "{{ packetsize }}",
            },
        },
        {
            "name": "protocol.enable",
            "getval": re.compile(
                r"""
                ^snmp-server
                \sprotocol\s(?P<enable>enable)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server protocol enable",
            "result": {
                "protocol": {
                    "enable": "{{ not not enable }}",
                },
            },
        },
        {
            "name": "source_interface.informs",
            "getval": re.compile(
                r"""
                ^snmp-server
                \ssource-interface\sinforms\s(?P<informs>\S+)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server source-interface informs {{ source_interface.informs }}",
            "result": {
                "source_interface": {
                    "informs": "{{ informs }}",
                },
            },
        },
        {
            "name": "source_interface.traps",
            "getval": re.compile(
                r"""
                ^snmp-server
                \ssource-interface\straps\s(?P<traps>\S+)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server source-interface traps {{ source_interface.traps }}",
            "result": {
                "source_interface": {
                    "traps": "{{ traps }}",
                },
            },
        },
        {
            "name": "system_shutdown",
            "getval": re.compile(
                r"""
                ^snmp-server
                \s(?P<system_shutdown>system-shutdown)
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server system-shutdown",
            "result": {
                "system_shutdown": "{{ not not system_shutdown }}",
            },
        },
        {
            "name": "tcp_session",
            "getval": re.compile(
                r"""
                ^snmp-server
                \s(?P<tcp_session>tcp-session)
                (\s(?P<auth>auth))?
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server tcp-session"
                      "{{ ' auth' if tcp_session.auth|d(False) else '' }}",
            "result": {
                "tcp_session": {
                    "enable": "{{ True if tcp_session is defined and auth is not defined else None }}",
                    "auth": "{{ not not auth }}",
                },
            },
        },
        {
            "name": "users.auth",
            "getval": re.compile(
                r"""
                ^snmp-server
                \suser\s(?P<user>\S+)
                (\s(?P<group>[^auth]\S+))?
                (\sauth\s(?P<algorithm>md5|sha|sha-256)\s(?P<password>\S+))?
                (\spriv(\s(?P<aes_128>aes-128))?\s(?P<privacy_password>\S+))?
                (\s(?P<localized_key>localizedkey))?
                (\s(?P<localizedv2_key>localizedV2key))?
                (\sengineID\s(?P<engine_id>\S+))?
                $""", re.VERBOSE,
            ),
            "setval": _tmplt_users_auth,
            "remval": "snmp-server user {{ user }}",
            "result": {
                "users": {
                    "auth": [
                        {
                            "user": "{{ user }}",
                            "group": "{{ group }}",
                            "authentication": {
                                "algorithm": "{{ algorithm }}",
                                "password": "'{{ password }}'",
                                "engine_id": "'{{ engine_id }}'",
                                "localized_key": "{{ not not localized_key }}",
                                "localizedv2_key": "{{ not not localizedv2_key }}",
                                "priv": {
                                    "privacy_password": "'{{ privacy_password }}'",
                                    "aes_128": "{{ not not aes_128 }}",
                                },
                            },
                        },
                    ],
                },
            },
        },
        {
            "name": "users.use_acls",
            "getval": re.compile(
                r"""
                ^snmp-server
                \suser\s(?P<user>\S+)
                (\suse-ipv4acl\s(?P<ipv4>\S+))?
                (\suse-ipv6acl\s(?P<ipv6>\S+))?
                $""", re.VERBOSE,
            ),
            "setval": "snmp-server user {{ user }}"
                      "{{ (' use-ipv4acl ' + ipv4) if ipv4 is defined else '' }}"
                      "{{ (' use-ipv6acl ' + ipv6) if ipv6 is defined else '' }}",
            "result": {
                "users": {
                    "use_acls": [
                        {
                            "user": "{{ user }}",
                            "ipv4": "{{ ipv4 }}",
                            "ipv6": "{{ ipv6 }}",
                        },
                    ],
                },
            },
        },
    ]
    # fmt: on



#!/usr/bin/python

# Copyright (c) Ansible project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: keycloak_realm_rolemapping

short_description: Allows administration of Keycloak realm role mappings into groups with the Keycloak API

version_added: 8.2.0

description:
  - This module allows you to add, remove or modify Keycloak realm role mappings into groups with the Keycloak REST API. It
    requires access to the REST API using OpenID Connect; the user connecting and the client being used must have the requisite
    access rights. In a default Keycloak installation, admin-cli and an admin user would work, as would a separate client
    definition with the scope tailored to your needs and a user having the expected roles.
  - The names of module options are snake_cased versions of the camelCase ones found in the Keycloak API and its documentation
    at U(https://www.keycloak.org/docs-api/18.0/rest-api/index.html).
  - Attributes are multi-valued in the Keycloak API. All attributes are lists of individual values and are returned that way
    by this module. You may pass single values for attributes when calling the module, and this is translated into a list
    suitable for the API.
  - When updating a group_rolemapping, where possible provide the role ID to the module. This removes a lookup to the API
    to translate the name into the role ID.
attributes:
  check_mode:
    support: full
  diff_mode:
    support: full
  action_group:
    version_added: 10.2.0

options:
  state:
    description:
      - State of the realm_rolemapping.
      - On C(present), the realm_rolemapping is created if it does not yet exist, or updated with the parameters you provide.
      - On C(absent), the realm_rolemapping is removed if it exists.
    default: 'present'
    type: str
    choices:
      - present
      - absent

  realm:
    type: str
    description:
      - They Keycloak realm under which this role_representation resides.
    default: 'master'

  group_name:
    type: str
    description:
      - Name of the group to be mapped.
      - This parameter is required (can be replaced by gid for less API call).
  parents:
    type: list
    description:
      - List of parent groups for the group to handle sorted top to bottom.
      - Set this if your group is a subgroup and you do not provide the GID in O(gid).
    elements: dict
    suboptions:
      id:
        type: str
        description:
          - Identify parent by ID.
          - Needs less API calls than using O(parents[].name).
          - A deep parent chain can be started at any point when first given parent is given as ID.
          - Note that in principle both ID and name can be specified at the same time but current implementation only always
            use just one of them, with ID being preferred.
      name:
        type: str
        description:
          - Identify parent by name.
          - Needs more internal API calls than using O(parents[].id) to map names to ID's under the hood.
          - When giving a parent chain with only names it must be complete up to the top.
          - Note that in principle both ID and name can be specified at the same time but current implementation only always
            use just one of them, with ID being preferred.
  gid:
    type: str
    description:
      - ID of the group to be mapped.
      - This parameter is not required for updating or deleting the rolemapping but providing it reduces the number of API
        calls required.
  roles:
    description:
      - Roles to be mapped to the group.
    type: list
    elements: dict
    suboptions:
      name:
        type: str
        description:
          - Name of the role_representation.
          - This parameter is required only when creating or updating the role_representation.
      id:
        type: str
        description:
          - The unique identifier for this role_representation.
          - This parameter is not required for updating or deleting a role_representation but providing it reduces the number
            of API calls required.
extends_documentation_fragment:
  - community.general._keycloak
  - community.general._keycloak.actiongroup_keycloak
  - community.general._attributes

author:
  - Gaëtan Daubresse (@Gaetan2907)
  - Marius Huysamen (@mhuysamen)
  - Alexander Groß (@agross)
"""

EXAMPLES = r"""
- name: Map a client role to a group, authentication with credentials
  community.general.keycloak_realm_rolemapping:
    realm: MyCustomRealm
    auth_client_id: admin-cli
    auth_keycloak_url: https://auth.example.com/auth
    auth_realm: master
    auth_username: USERNAME
    auth_password: PASSWORD
    state: present
    group_name: group1
    roles:
      - name: role_name1
        id: role_id1
      - name: role_name2
        id: role_id2
  delegate_to: localhost

- name: Map a client role to a group, authentication with token
  community.general.keycloak_realm_rolemapping:
    realm: MyCustomRealm
    auth_client_id: admin-cli
    auth_keycloak_url: https://auth.example.com/auth
    token: TOKEN
    state: present
    group_name: group1
    roles:
      - name: role_name1
        id: role_id1
      - name: role_name2
        id: role_id2
  delegate_to: localhost

- name: Map a client role to a subgroup, authentication with token
  community.general.keycloak_realm_rolemapping:
    realm: MyCustomRealm
    auth_client_id: admin-cli
    auth_keycloak_url: https://auth.example.com/auth
    token: TOKEN
    state: present
    group_name: subgroup1
    parents:
      - name: parent-group
    roles:
      - name: role_name1
        id: role_id1
      - name: role_name2
        id: role_id2
  delegate_to: localhost

- name: Unmap realm role from a group
  community.general.keycloak_realm_rolemapping:
    realm: MyCustomRealm
    auth_client_id: admin-cli
    auth_keycloak_url: https://auth.example.com/auth
    auth_realm: master
    auth_username: USERNAME
    auth_password: PASSWORD
    state: absent
    group_name: group1
    roles:
      - name: role_name1
        id: role_id1
      - name: role_name2
        id: role_id2
  delegate_to: localhost
"""

RETURN = r"""
msg:
  description: Message as to what action was taken.
  returned: always
  type: str
  sample: "Role role1 assigned to group group1."

proposed:
  description: Representation of proposed client role mapping.
  returned: always
  type: dict
  sample: {"clientId": "test"}

existing:
  description:
    - Representation of existing client role mapping.
    - The sample is truncated.
  returned: always
  type: dict
  sample:
    {
      "adminUrl": "http://www.example.com/admin_url",
      "attributes": {
        "request.object.signature.alg": "RS256"
      }
    }

end_state:
  description:
    - Representation of client role mapping after module execution.
    - The sample is truncated.
  returned: on success
  type: dict
  sample:
    {
      "adminUrl": "http://www.example.com/admin_url",
      "attributes": {
        "request.object.signature.alg": "RS256"
      }
    }
"""

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.general.plugins.module_utils._keycloak import (
    KeycloakAPI,
    KeycloakError,
    get_token,
    keycloak_argument_spec,
)


def main():
    """
    Module execution

    :return:
    """
    argument_spec = keycloak_argument_spec()

    roles_spec = dict(
        name=dict(type="str"),
        id=dict(type="str"),
    )

    meta_args = dict(
        state=dict(default="present", choices=["present", "absent"]),
        realm=dict(default="master"),
        gid=dict(type="str"),
        group_name=dict(type="str"),
        parents=dict(
            type="list",
            elements="dict",
            options=dict(id=dict(type="str"), name=dict(type="str")),
        ),
        roles=dict(type="list", elements="dict", options=roles_spec),
    )

    argument_spec.update(meta_args)

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_one_of=(
            [["token", "auth_realm", "auth_username", "auth_password", "auth_client_id", "auth_client_secret"]]
        ),
        required_together=([["auth_username", "auth_password"]]),
        required_by={"refresh_token": "auth_realm"},
    )

    result = dict(changed=False, msg="", diff={}, proposed={}, existing={}, end_state={})

    # Obtain access token, initialize API
    try:
        connection_header = get_token(module.params)
    except KeycloakError as e:
        module.fail_json(msg=str(e))

    kc = KeycloakAPI(module, connection_header)

    realm = module.params.get("realm")
    state = module.params.get("state")
    gid = module.params.get("gid")
    group_name = module.params.get("group_name")
    roles = module.params.get("roles")
    parents = module.params.get("parents")

    # Check the parameters
    if gid is None and group_name is None:
        module.fail_json(msg="Either the `group_name` or `gid` has to be specified.")

    # Get the potential missing parameters
    if gid is None:
        group_rep = kc.get_group_by_name(group_name, realm=realm, parents=parents)
        if group_rep is not None:
            gid = group_rep["id"]
        else:
            module.fail_json(msg=f"Could not fetch group {group_name}:")
    else:
        group_rep = kc.get_group_by_groupid(gid, realm=realm)

    if roles is None:
        module.exit_json(msg="Nothing to do (no roles specified).")
    else:
        for role in roles:
            if role["name"] is None and role["id"] is None:
                module.fail_json(msg="Either the `name` or `id` has to be specified on each role.")
            # Fetch missing role_id
            if role["id"] is None:
                role_rep = kc.get_realm_role(role["name"], realm=realm)
                if role_rep is not None:
                    role["id"] = role_rep["id"]
                else:
                    module.fail_json(msg=f"Could not fetch realm role {role['name']} by name:")
            # Fetch missing role_name
            else:
                for realm_role in kc.get_realm_roles(realm=realm):
                    if realm_role["id"] == role["id"]:
                        role["name"] = realm_role["name"]
                        break

                if role["name"] is None:
                    module.fail_json(msg=f"Could not fetch realm role {role['id']} by ID")

    assigned_roles_before = group_rep.get("realmRoles", [])

    result["existing"] = assigned_roles_before
    result["proposed"] = list(assigned_roles_before) if assigned_roles_before else []

    update_roles = []
    for role in roles:
        # Fetch roles to assign if state present
        if state == "present":
            if any(assigned == role["name"] for assigned in assigned_roles_before):
                pass
            else:
                update_roles.append(
                    {
                        "id": role["id"],
                        "name": role["name"],
                    }
                )
                result["proposed"].append(role["name"])
        # Fetch roles to remove if state absent
        else:
            if any(assigned == role["name"] for assigned in assigned_roles_before):
                update_roles.append(
                    {
                        "id": role["id"],
                        "name": role["name"],
                    }
                )
                if role["name"] in result["proposed"]:  # Handle double removal
                    result["proposed"].remove(role["name"])

    if len(update_roles):
        result["changed"] = True
        if module._diff:
            result["diff"] = dict(before=assigned_roles_before, after=result["proposed"])
        if module.check_mode:
            module.exit_json(**result)

        if state == "present":
            # Assign roles
            kc.add_group_realm_rolemapping(gid=gid, role_rep=update_roles, realm=realm)
            result["msg"] = f"Realm roles {update_roles} assigned to groupId {gid}."
        else:
            # Remove mapping of role
            kc.delete_group_realm_rolemapping(gid=gid, role_rep=update_roles, realm=realm)
            result["msg"] = f"Realm roles {update_roles} removed from groupId {gid}."

        if gid is None:
            assigned_roles_after = kc.get_group_by_name(group_name, realm=realm, parents=parents).get("realmRoles", [])
        else:
            assigned_roles_after = kc.get_group_by_groupid(gid, realm=realm).get("realmRoles", [])
        result["end_state"] = assigned_roles_after
        module.exit_json(**result)
    # Do nothing
    else:
        result["changed"] = False
        result["msg"] = (
            f"Nothing to do, roles {roles} are {'mapped' if state == 'present' else 'not mapped'} with group {group_name}."
        )
        module.exit_json(**result)


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or
# https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
module: organizations_inventory_claim
short_description: Resource module for organizations _inventory _claim
description:
  - Manage operation create of the resource organizations _inventory _claim. - > Claim
    a list of devices, licenses, and/or orders into an organization inventory. When
    claiming by order, all devices and licenses in the order will be claimed; licenses
    will be added to the organization and devices will be placed in the organization's
    inventory. This operation can be used up to ten times within a single five minute
    window.
version_added: '1.0.0'
extends_documentation_fragment:
  - cisco.meraki.module
author: Francisco Munoz (@fmunoz)
options:
  licenses:
    description: The licenses that should be claimed.
    elements: dict
    suboptions:
      key:
        description: The key of the license.
        type: str
      mode:
        description: Co-term licensing only either 'renew' or 'addDevices'. 'addDevices'
          will increase the license limit, while 'renew' will extend the amount of
          time until expiration. Defaults to 'addDevices'. All licenses must be claimed
          with the same mode, and at most one renewal can be claimed at a time. Does
          not apply to organizations using per-device licensing model.
        type: str
    type: list
  orders:
    description: The numbers of the orders that should be claimed.
    elements: str
    type: list
  organizationId:
    description: OrganizationId path parameter. Organization ID.
    type: str
  serials:
    description: The serials of the devices that should be claimed.
    elements: str
    type: list
requirements:
  - meraki >= 2.4.9
  - python >= 3.5
seealso:
  - name: Cisco Meraki documentation for organizations claimIntoOrganizationInventory
    description: Complete reference of the claimIntoOrganizationInventory API.
    link: https://developer.cisco.com/meraki/api-v1/#!claim-into-organization-inventory
notes:
  - SDK Method used are
    organizations.Organizations.claim_into_organization_inventory,
  - Paths used are
    post /organizations/{organizationId}/inventory/claim,
"""

EXAMPLES = r"""
- name: Create
  cisco.meraki.organizations_inventory_claim:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    licenses:
      - key: Z2XXXXXXXXXX
        mode: addDevices
    orders:
      - 4CXXXXXXX
    organizationId: string
    serials:
      - Q234-ABCD-5678
"""
RETURN = r"""
meraki_response:
  description: A dictionary or list with the response returned by the Cisco Meraki Python SDK
  returned: always
  type: dict
  sample: >
    {
      "orders": [
        "string"
      ],
      "serials": [
        "string"
      ],
      "licenses": [
        {
          "key": "string",
          "mode": "string"
        }
      ]
    }
"""
